# THIS-FILE-WILL-BE-REPLACED !!! DO NOT CHANGE OR MODIFY THIS FILE !!!

# During `setup.py build_py`, this file is overwritten
# __version__ = '[git describe --abbrev=7 --tags --dirty=+dirty --always]
# __build__ = '[git rev-parse HEAD]'

# PEP 440 -- Version Identification and Dependency Specification
# https://www.python.org/dev/peps/pep-0440/


__version__ = 'devel'
__build__ = 'devel'



import aiohttp.abc

from ..log import LOG_NOTICE


class AccessLogger(aiohttp.abc.AbstractAccessLogger):

	def log(self, request, response, time):
		struct_data = {
			'I': request.remote,
			'al.m': request.method,
			'al.p': request.path,
			'al.c': response.status,
			'D': time,
		}

		if request.content_length is not None:
			struct_data['al.B'] = request.content_length

		if response.content_length is not None:
			struct_data['al.b'] = response.content_length

		if hasattr(request, 'Identity'):
			struct_data['i'] = request.Identity

		agent = request.headers.get('User-Agent')
		if agent is not None:
			struct_data['al.A'] = agent

		xfwd = request.headers.get('X-Forwarded-For')
		if xfwd is not None:
			# TODO: Sanitize xfwd
			# In nginx, use "proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;"
			struct_data['Ix'] = xfwd[:128]

		self.logger.log(LOG_NOTICE, '', struct_data=struct_data)



import aiohttp

from ...abc.service import Service
from ...config import Config

from .tenant import Tenant

# "tenant_url" is used to periodically refresh tenants from, expecting "_id" inside a JSON structure,
# which is compatible with SeaCat Auth product
Config.add_defaults({
	'tenants': {
		'ids': '', # List of tenant ids, entries can be separated by comma or newline
		'tenant_url': '',  # f. e. http://seacat-auth:8080/tenant
		'trusted': 0,  # makes sure the tenants are implicitly trusted, even though they are not located in IDs or tenant URL
	}
})


class TenantService(Service):

	def __init__(self, app, service_name="asab.TenantService"):
		super().__init__(app, service_name)
		self.App = app
		self.TenantWebHandler = None
		self.TenantsTrusted = int(Config['tenants']['trusted'])
		self.Tenants = {}

		# Load tenants from configuration
		self.TenantIds = Config['tenants']['ids']
		self.TenantIds = self.TenantIds.replace(',', '\n').split('\n')

		for tenant_id in self.TenantIds:
			tenant_id = tenant_id.strip()
			if len(tenant_id) == 0: continue
			if tenant_id[0] == '#': continue
			if tenant_id[0] == ';': continue

			section = 'tenant:params:{}'.format(tenant_id)
			if Config.has_section(section):
				params = dict(Config.items(section))
				self.Tenants[tenant_id] = Tenant(tenant_id, params)
			else:
				self.Tenants[tenant_id] = Tenant(tenant_id)

		# Load tenants from URL
		self.TenantUrl = Config["tenants"]["tenant_url"]


	async def initialize(self, app):
		if len(self.TenantUrl) > 0:
			await self._update_tenants()
			# TODO: Websocket persistent API should be added to seacat auth to feed these changes in realtime (eventually)
			app.PubSub.subscribe("Application.tick/300!", self._update_tenants)


	def locate_tenant(self, tenant_id):
		tenant = self.Tenants.get(tenant_id)
		if tenant is None and self.TenantsTrusted > 0:
			tenant = {"_id": tenant_id}
			self.Tenants[tenant_id] = tenant
		return tenant


	def get_tenants(self):
		tenants = []
		for tenant in self.Tenants.values():
			tenants.append(tenant.to_dict())
		return tenants


	def add_web_api(self, web_container):
		from .web import TenantWebHandler
		self.TenantWebHandler = TenantWebHandler(self.App, self, web_container)


	async def _update_tenants(self, message_name=None):
		async with aiohttp.ClientSession() as session:
			async with session.get(self.TenantUrl) as resp:
				if resp.status == 200:
					tenants_list = await resp.json()
					for tenant in tenants_list:
						self.Tenants[tenant["_id"]] = tenant



import aiozk
import asyncio
import json

from ..config import ConfigObject


class ZooKeeperContainer(ConfigObject):
	"""
	ZooKeeperContainer connects to Zookeeper via aiozk client:
	https://zookeeper.apache.org/
	https://pypi.org/project/aiozk/
	"""

	ConfigDefaults = {
		"urls": "zookeeper:12181",
		"path": "/asab",
	}

	def __init__(self, app, config_section_name, config=None):
		super().__init__(config_section_name=config_section_name, config=config)
		self.App = app
		self.ConfigSectionName = config_section_name
		self.ZooKeeper = aiozk.ZKClient(self.Config["urls"])
		self.ZooKeeperPath = self.Config["path"]

	async def initialize(self, app):
		await self.ZooKeeper.start()
		await self.ZooKeeper.ensure_path(self.ZooKeeperPath)

	async def finalize(self, app):
		await self.ZooKeeper.close()

	async def advertise(self, data, encoding="utf-8"):
		if isinstance(data, dict):
			data = json.dumps(data).encode(encoding)
		elif isinstance(data, str):
			data = data.encode(encoding)
		elif asyncio.iscoroutinefunction(data):
			data = await data
		elif callable(data):
			data = data()

		return await self.ZooKeeper.create(
			"{}/i".format(self.ZooKeeperPath),
			data=data,
			sequential=True,
			ephemeral=True
		)

	async def get_children(self):
		return await self.ZooKeeper.get_children(self.ZooKeeperPath)

	async def get_data(self, child, encoding="utf-8"):
		raw_data = await self.get_raw_data(child)
		if raw_data is None:
			return {}
		return json.loads(raw_data.decode(encoding))

	async def get_raw_data(self, child):
		return await self.ZooKeeper.get_data("{}/{}".format(self.ZooKeeperPath, child))



import asab
import concurrent.futures


class ProactorService(asab.Service):

	def __init__(self, app, service_name):
		super().__init__(app, service_name)
		self.Loop = app.Loop

		max_workers = asab.Config.get('asab:proactor', 'max_workers')
		try:
			max_workers = int(max_workers)
		except BaseException:
			max_workers = None
		if max_workers <= 0:
			max_workers = None

		self.Executor = concurrent.futures.ThreadPoolExecutor(
			max_workers=max_workers,
			thread_name_prefix="AsabProactorThread"
		)

		if asab.Config.get('asab:proactor', 'default_executor'):
			self.Loop.set_default_executor(self.Executor)


	# There was the method run, which is obsolete
	def execute(self, func, *args):
		return self.Loop.run_in_executor(self.Executor, func, *args)





class Tenant(object):

	def __init__(self, tenant_id: str, params: dict = {}):

		self.Id = tenant_id
		self.Parameters = params

	def to_dict(self):

		return {
			'id': self.Id,
			'parameters': self.Parameters
		}



import asab


class MetricsMemstorTarget(asab.ConfigObject):

	def __init__(self, svc, config_section_name, config=None):
		super().__init__(config_section_name, config)
		self.Metrics = {}


	async def process(self, now, mlist):

		for metric, values in mlist:
			# Build metric name with tags
			name = metric.Name
			for tk, tv in metric.Tags.items():
				name += ',{}={}'.format(tk, tv)
			# Store metric
			self.Metrics[name] = {
				"Timestamp": now,
				"Name": metric.Name,
				"Values": values,
				"Tags": metric.Tags
			}


	def rest_get(self):
		return self.Metrics



import urllib.parse
import hashlib
import os.path
import glob
import logging

import aiohttp.web

import cryptography.x509
import cryptography.hazmat.backends
import cryptography.hazmat.primitives.hashes
import cryptography.hazmat.primitives.serialization

from ...pdict import PersistentDict
from ...config import ConfigObject
from ... import Service

#

L = logging.getLogger(__name__)

#


def pubkeyauth_middleware_factory(app, *args, mode='direct', service=None, **kwargs):
	'''
	`service` is the instance of `PublicKeyAuthenticationService`.
	If `service` is not provided, every SSL client with certificate is authenticated.

	`mode` is one of `direct` or `proxy`.

	'''

	@aiohttp.web.middleware
	async def pubkeyauth_direct_middleware(request, handler):
		'''
		This middleware is used when ASAB is working directly with SSL socket
		'''

		ssl_object = request.transport.get_extra_info('ssl_object')
		if ssl_object is None:
			# The connection is not a SSL
			return await handler(request)

		cert = ssl_object.getpeercert(binary_form=True)
		if cert is None:
			# The client doesn't provided a certificate
			return await handler(request)

		try:
			cert = cryptography.x509.load_der_x509_certificate(
				cert,
				cryptography.hazmat.backends.default_backend()
			)
		except Exception:
			L.exception("Error when parsing a client certificate")
			return await handler(request)


		if service is not None:
			if not service.authenticate(cert):
				return await handler(request)

		request.Identity = cert.subject.rfc4514_string()

		return await handler(request)

	return pubkeyauth_direct_middleware


	@aiohttp.web.middleware
	async def pubkeyauth_proxy_middleware(request, handler):
		'''
		This middleware is used when ASAB is working behind a SSL-terminating proxy.
		Client certificate is expected in X-SSL-Client-Cert

		The client certificate is now extracted from `X-SSL-Client-Cert` header, where is it stored by NGinx by:

		proxy_set_header X-SSL-Client-Cert $ssl_client_escaped_cert;

server {
	listen  443 ssl;

	# A server certificate
	ssl_certificate_key letsencrypt/key.pem;
	ssl_certificate	letsencrypt/fullchain.cer;

	# Certificate of your custom CA for clients
	ssl_trusted_certificate custom-ca-cert.pem;

	# make verification optional, so we can display a 403 message to those who fail authentication
	ssl_verify_client optional_no_ca;

	location /websocket {
		if ($ssl_client_verify != SUCCESS) {
			return 403;
		}

		proxy_pass http://localhost:8080/websocket;
		proxy_http_version 1.1;

		proxy_set_header Host $host;
		proxy_set_header X-Real-IP $remote_addr;
		proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
		proxy_set_header X-Forwarded-Proto $scheme;

		proxy_set_header Upgrade $http_upgrade;
		proxy_set_header Connection "Upgrade";

		proxy_set_header X-Forwarded-Client-Cert $ssl_client_escaped_cert;
	}
}

		'''

		cert = request.headers.get('X-Forwarded-Client-Cert')
		if cert is None:
			return await handler(request)
		cert = urllib.parse.unquote_to_bytes(cert)

		try:
			cert = cryptography.x509.load_pem_x509_certificate(
				cert,
				cryptography.hazmat.backends.default_backend()
			)
		except Exception:
			L.exception("Error when parsing a client certificate")
			return await handler(request)

		if service is not None:
			if not service.authenticate(cert):
				return await handler(request)

		request.Identity = cert.subject.rfc4514_string()

		return await handler(request)

	if mode == 'direct':
		return pubkeyauth_direct_middleware
	elif mode == 'proxy':
		return pubkeyauth_proxy_middleware
	else:
		raise RuntimeError("Unknown mode '{}'".format(mode))



class PublicKeyAuthenticationService(Service, ConfigObject):

	'''
	This is an authentication service that uses the whitelist of public keys of authenticated clients.
	Clients should provide a certificate via TLS handshake (aka mutual TLS authorization).
	The public key from a certificate is then matched with certificates that are stored in the directory cache.
	A client is authenticated when the matching certificate is found, otherwise "HTTPUnauthorized" (401) is raised.

	Client certificates are issued by a dedicated Certificate Authority. You can establish your own.
	A public certificate of the client has to be placed into a client certificate directory of the server.

	This authentication is designed from machine-to-machine websocket communication with small amount of requests.
	It validates the certificate/public keys every time the client hits the server.
	That is OK for long-lived WebSockets, but not scalable for a regular HTTP traffic.

	'''

	ConfigDefaults = {
		'dir': './',
		'glob': '*-cert.pem',
		'index': '.index.bin',
	}

	def __init__(self, app, *, service_name="asab.PublicKeyAuthenticationService", config_section_name='asab:web:authn:pubkey', config=None):
		super().__init__(app, service_name)
		ConfigObject.__init__(self, config_section_name, config)

		self.ClientCertDir = self.Config.get('dir')
		self.ClientCertGlob = self.Config.get('glob')
		self.IndexPDict = PersistentDict(os.path.join(self.ClientCertDir, self.Config.get('index')))


	def authenticate(self, certificate):
		public_key = certificate.public_key()
		pk_digest = self.get_public_key_digest(public_key)

		entry = self.IndexPDict.get(pk_digest)
		if entry is None:
			# Key not found in the index, let's scan the directory
			self._scan_dir()
			entry = self.IndexPDict.get(pk_digest)

		if entry is None:
			L.warning("Authentication failed, public key not found")
			return False

		assert(entry is not None)

		pk_digest1 = self.get_public_key_digest_from_filename(entry)
		if pk_digest1 is None:
			return False

		return pk_digest == pk_digest1


	def _scan_dir(self):
		known_fnames = frozenset(self.IndexPDict.values())
		for fname in glob.glob(os.path.join(self.ClientCertDir, self.ClientCertGlob)):
			if fname in known_fnames:
				continue
			pk_digest = self.get_public_key_digest_from_filename(fname)
			self.IndexPDict[pk_digest] = fname


	def get_public_key_digest_from_filename(self, fname):
		try:
			with open(fname, 'rb') as f:
				# Load a client certificate in PEM format
				cert = cryptography.x509.load_pem_x509_certificate(
					f.read(),
					cryptography.hazmat.backends.default_backend()
				)
		except Exception:
			return None

		return self.get_public_key_digest(cert.public_key())


	def get_public_key_digest(self, public_key):
		# Hash the public key
		public_key_bytes = public_key.public_bytes(
			cryptography.hazmat.primitives.serialization.Encoding.DER,
			cryptography.hazmat.primitives.serialization.PublicFormat.SubjectPublicKeyInfo
		)
		h = hashlib.blake2b(public_key_bytes)
		return h.digest()



from .abc.module import Module
from .abc.service import Service
from .abc.singleton import Singleton
from .application import Application
from .config import Config, ConfigObject
from .log import LOG_NOTICE
from .pdict import PersistentDict
from .pubsub import subscribe, PubSub, Subscriber
from .socket import StreamSocketServerService
from .timer import Timer

from .__version__ import __version__, __build__


__all__ = (
	'Module',
	'Service',
	'Singleton',
	'Application',
	'Config',
	'ConfigObject',
	'LOG_NOTICE',
	'PersistentDict',
	'subscribe',
	'PubSub',
	'Subscriber',
	'StreamSocketServerService',
	'Timer',
	'__version__',
	'__build__',
)



import functools

import aiohttp
import aiohttp.web


def authn_middleware_factory(app, implementation, *args, **kwargs):

	if implementation == "oauth2client":
		from .oauth import oauthclient_middleware_factory
		return oauthclient_middleware_factory(app, *args, **kwargs)

	elif implementation == "basicauth":
		from .basicauth import basicauth_middleware_factory
		return basicauth_middleware_factory(app, *args, **kwargs)

	elif implementation == "pubkeyauth":
		from .pubkeyauth import pubkeyauth_middleware_factory
		return pubkeyauth_middleware_factory(app, *args, **kwargs)

	else:
		raise RuntimeError("Unknown authentication implementation '{}'".format(implementation))


def authn_required_handler(func):
	'''
	The web handler method require the authentication.
	The peer identity is available in the `identity` keywork argument.

	Example of use:

	@asab.web.authn.authn_required_handler
	async def endpoint(self, request, *, identity):
		...
	'''

	async def wrapper(*args, **kargs):
		request = args[-1]

		try:
			kargs['identity'] = request.Identity
		except AttributeError:
			raise aiohttp.web.HTTPUnauthorized()

		return await func(*args, **kargs)

	return wrapper


def authn_optional_handler(func):
	'''
	The web handler method request the OPTINAL authentication.
	The peer identity is available in the `identity` keywork argument.
	`identity` is None if the peer has not provided authentication info.

	Example of use:

	@asab.web.authn.authn_optional_handler
	async def endpoint(self, request, *, identity):
		...
	'''

	async def wrapper(*args, **kargs):
		request = args[-1]

		try:
			kargs['identity'] = request.Identity
		except AttributeError:
			kargs['identity'] = None

		return await func(*args, **kargs)

	return wrapper


def authorize_any(*authorizations):
	"""
	The web handler that checks if ANY *authorizations* passed successfully.

	Example:
		@asab.web.authn.authorize_any(my_custom_authorization, allow_admin)
		async def handler(self, request, *):
			...

	:param authorizations: List of authorization functions
	"""
	def decorator_authorize_any(func):

		@functools.wraps(func)
		async def wrapper_authorize_any(*args, **kwargs):
			result = [await auth(args, kwargs) for auth in authorizations]
			if any(result):
				return await func(*args, **kwargs)
			else:
				raise aiohttp.web.HTTPForbidden()

		return wrapper_authorize_any

	return decorator_authorize_any


def authorize_all(*authorizations):
	"""
	The web handler that checks if ALL *authorizations* passed successfully.

	Example:
		@asab.web.authn.authorize_all(my_custom_authorization, allow_admin)
		async def handler(self, request, *):
			...

	:param authorizations: List of authorization functions
	"""
	def decorator_authorize_all(func):

		@functools.wraps(func)
		async def wrapper_authorize_all(*args, **kwargs):
			result = [await auth(args, kwargs) for auth in authorizations]
			if all(result):
				return await func(*args, **kwargs)
			else:
				raise aiohttp.web.HTTPForbidden()

		return wrapper_authorize_all

	return decorator_authorize_all



import uuid
import logging

from .storage import SessionStorage
from .cookies import CookieSessionMixIn


L = logging.getLogger(__name__)


class InMemorySessionStorage(SessionStorage, CookieSessionMixIn):


	def __init__(self, app, session_class, max_age=60 * 15, cookie_name='SESSID'):
		super().__init__(app, max_age=max_age, session_class=session_class)
		self.set_cookie_name(cookie_name)

		self.Sessions = {}

		app.PubSub.subscribe("Application.tick/10!", self._on_tick)


	async def load(self, request):
		session_id = self.get_session_id_from_request(request)
		if session_id is not None:
			session = self.Sessions.get(session_id)
			if session is not None:
				if session.is_expired():
					session = None
				else:
					return session

			L.warning("Invalid session id '{}'".format(session_id))

		return await self.create(request)


	async def set(self, session, response):
		if session.is_new():

			while True:
				session_id = uuid.uuid4().hex
				if session_id not in self.Sessions:
					session.set_id(session_id)
					break

			self.Sessions[session.Id] = session
			self.set_session_to_response(session, response)


	async def delete(self, session):
		session_p = self.Sessions.pop(session.Id)
		assert session_p == session


	async def store(self, session):
		session.reset()


	async def _on_tick(self, message_type):
		# Find expired session and remove them
		expired_sessions = []
		for session in self.Sessions.values():
			if session.is_expired():
				expired_sessions.append(session)

		for session in expired_sessions:
			await self.delete(session)



import os
import asab
import asab.web
import asab.web.rest
from asab.api.log import WebApiLoggingHandler
import logging

##

L = logging.getLogger(__name__)


##


class ApiService(asab.Service):

	def __init__(self, app, service_name):
		super().__init__(app, service_name)

		listen = asab.Config["asab:web"]["listen"]
		self.Container = self._initialize_web(app, listen)


	def _initialize_web(self, app, listen):
		websvc = app.get_service("asab.WebService")

		# Create a dedicated web container
		container = asab.web.WebContainer(
			websvc, "asab:web",
			config={"listen": listen}
		)
		# TODO: refactor to use custom config section, instead of explicitly passing "listen" param?

		# TODO: Logging level configurable via config file
		self.APILogHandler = WebApiLoggingHandler(app, level=logging.NOTSET)
		self.format = logging.Formatter("%%(asctime)s %%(levelname)s %%(name)s %%(struct_data)s%%(message)s")
		self.APILogHandler.setFormatter(self.format)
		self.Logging = logging.getLogger()
		self.Logging.addHandler(self.APILogHandler)

		# Add routes
		container.WebApp.router.add_get('/asab/v1/environ', self.environ)
		container.WebApp.router.add_get('/asab/v1/config', self.config)

		container.WebApp.router.add_get('/asab/v1/logs', self.APILogHandler.get_logs)
		container.WebApp.router.add_get('/asab/v1/logws', self.APILogHandler.ws)

		return container


	async def environ(self, request):
		return asab.web.rest.json_response(request, dict(os.environ))


	async def config(self, request):
		# Copy the config and erase all passwords
		result = {}
		for section in asab.Config.sections():
			result[section] = {}
			# Access items in the raw mode (they are not interpolated)
			for option, value in asab.Config.items(section, raw=True):
				if section == "passwords":
					result[section][option] = "***"
				else:
					result[section][option] = value
		return asab.web.rest.json_response(request, result)



import logging
import aiohttp.web

import asab

from .session import Session

#

L = logging.getLogger(__name__)

#

# TODO: Configure cookie name
# TODO: Configure cookie http only / secure
# TODO: Configure cookie domain
# TODO: Configure max-age


class ServiceWebSession(asab.Service):

	def __init__(self, app, service_name, webapp, session_storage=None, session_class=None):
		super().__init__(app, service_name)

		if session_class is None:
			session_class = Session

		# Construct session storage
		if session_storage is None:
			from .inmemstor import InMemorySessionStorage
			self.SessionStorage = InMemorySessionStorage(app, session_class)
		else:
			self.SessionStorage = session_storage

		# Add middleware to a webservice
		webapp.middlewares.append(
			session_middleware(
				self.SessionStorage
			)
		)


def session_middleware(storage):

	@aiohttp.web.middleware
	async def factory(request, handler):
		session = await storage.load(request)
		request['Session'] = session
		try:
			response = await handler(request)
			if session.Id is None:
				await storage.set(session, response)
			return response
		finally:
			await storage.store(session)

	return factory



import abc
import asab


class StorageServiceABC(asab.Service):

	def __init__(self, app, service_name):
		super().__init__(app, service_name)


	@abc.abstractmethod
	def upsertor(self, collection: str, obj_id=None, version: int = 0):
		'''
		If updating an existing object, please specify its `obj_id` and also `version` that you need to read from a storage upfront.
		If `obj_id` is None, we assume that you want to insert a new object and generate its new `obj_id`, `version` should be set to 0 (default) in that case.
		If you want to insert a new object with a specific `obj_id`, specify `obj_id` and set a version to 0.
			- If there will be a colliding object already stored in a storage, `execute()` method will fail on `DuplicateError`.

		:param collection: Name of collection to work with
		:param obj_id: Primary identification of an object in the storage (e.g. primary key)
		:param version: Specify a current version of the object and hence prevent byzantine faults. \
						You should always read the version from the storage upfront, prior using an upsertor. \
						That creates a soft lock on the record. It means that if the object is updated by other \
						component in meanwhile, your upsertor will fail and you should retry the whole operation. \
						The new objects should have a `version` set to 0.
		'''
		pass


	@abc.abstractmethod
	async def get(self, collection: str, obj_id) -> dict:
		"""
		Get object from collection

		:param collection: Collection to get from
		:param obj_id: Object identification
		:return: The object retrieved from a storage

		Raises:
			KeyError: If `obj_id` not found in `collection`
		"""
		pass


	@abc.abstractmethod
	async def get_by(self, collection: str, key: str, value):
		pass


	@abc.abstractmethod
	async def delete(self, collection: str, obj_id):
		pass



import asyncio
import logging
import datetime

import aiohttp

from ..web.rest.json import json_response
from ..log import LOG_NOTICE

##

L = logging.getLogger(__name__)


##


class WebApiLoggingHandler(logging.Handler):


	def __init__(self, app, level=logging.NOTSET, buffer_size: int = 10):
		super().__init__(level=level)

		self.Buffer = []
		self._buffer_size = buffer_size
		self.WebSockets = set()

		app.PubSub.subscribe("Application.stop!", self._on_stop)


	async def _on_stop(self, _on_stop, x):
		for ws in list(self.WebSockets):
			await ws.send_json({
				"t": datetime.datetime.utcnow().isoformat() + 'Z',
				"C": "asab.web",
				"M": "Closed.",
				"l": logging.INFO,
			})
			await ws.close()


	def emit(self, record):
		if logging.DEBUG < record.levelno <= logging.INFO:
			severity = 6  # Informational
		elif record.levelno <= LOG_NOTICE:
			severity = 5  # Notice
		elif record.levelno <= logging.WARNING:
			severity = 4  # Warning
		elif record.levelno <= logging.ERROR:
			severity = 3  # Error
		elif record.levelno <= logging.CRITICAL:
			severity = 2  # Critical
		else:
			severity = 1  # Alert

		log_entry = {
			"t": datetime.datetime.utcfromtimestamp(record.created).isoformat() + 'Z',
			"C": record.name,
			"s": "{}:{}".format(record.funcName, record.lineno),
			"p": record.process,
			"Th": record.thread,
			"l": severity,
		}

		message = record.getMessage()
		if record.exc_text is not None:
			message += '\n' + record.exc_text
		if record.stack_info is not None:
			message += '\n' + record.stack_info
		if len(message) > 0:
			log_entry['M'] = message

		sd = record.__dict__.get("_struct_data")
		if sd is not None:
			log_entry['sd'] = sd

		if len(self.Buffer) > self._buffer_size:
			del self.Buffer[0]
			self.Buffer.append(log_entry)

		else:
			self.Buffer.append(log_entry)

		if len(self.WebSockets) > 0:
			asyncio.ensure_future(self._send_ws(log_entry))


	async def get_logs(self, request):
		return json_response(request, self.Buffer)


	async def ws(self, request):
		'''
		Websocket connection

		Usable with e.g. with React Lazylog
		https://github.com/mozilla-frontend-infra/react-lazylog#readme

		<LazyLog
			url={this.AsabLogWsURL}
			follow
			websocket
			websocketOptions={{
				formatMessage: e => log_message_format(e),
			}}
		/>

		function log_message_format(e) {
			e = JSON.parse(e)
			var msg = e.t;
			if (e.l != undefined) msg += " " + e.l;
			if (e.sd != undefined) msg += ` ${JSON.stringify(e.sd)}`;
			if (e.M != undefined) msg += " " + e.M;
			if (e.C != undefined) msg += ` [${e.C}]`
			return msg;
		}

		'''

		ws = aiohttp.web.WebSocketResponse()
		await ws.prepare(request)

		await ws.send_json({
			"t": datetime.datetime.utcnow().isoformat() + 'Z',
			"C": "asab.web",
			"M": "Connected.",
			"l": logging.INFO,
		})

		# Send historical logs
		for log_entry in self.Buffer:
			await ws.send_json(log_entry)

		self.WebSockets.add(ws)
		try:

			async for msg in ws:
				if msg.type == aiohttp.WSMsgType.ERROR:
					break

		finally:
			self.WebSockets.remove(ws)

		return ws


	async def _send_ws(self, log_entry):
		for ws in self.WebSockets:
			await ws.send_json(log_entry)



import asyncio
import asab


class WebService(asab.Service):

	def __init__(self, app, service_name):
		super().__init__(app, service_name)

		self.Containers = {}
		self.App = app


	async def finalize(self, app):
		for containers in self.Containers.values():
			await containers.finalize(app)


	def _register_container(self, container, config_section_name):
		self.Containers[config_section_name] = container
		asyncio.ensure_future(container.initialize(self.App))


	@property
	def WebApp(self):
		'''
		This is here to maintain backward compatibility.
		'''
		try:
			return self.Containers['asab:web'].WebApp
		except KeyError:
			from .container import WebContainer
			return WebContainer(self, 'asab:web').WebApp



import asab

from .service import WebService
from .websocket import WebSocketFactory
from .staticdir import StaticDirProvider
from .container import WebContainer


class Module(asab.Module):

	def __init__(self, app):
		super().__init__(app)
		self.service = WebService(app, "asab.WebService")


__all__ = (
	'WebService',
	'WebSocketFactory',
	'StaticDirProvider',
	'WebContainer',
	'Module',
)



import logging
import asyncio
import argparse
import itertools
import os
import sys
import time
import signal
import platform
import random

try:
	import daemon
	import daemon.pidfile
	import lockfile
except ImportError:
	daemon = None

from .config import Config
from .abc.singleton import Singleton
from .log import Logging, _loop_exception_handler


L = logging.getLogger(__name__)


class Application(metaclass=Singleton):

	Description = "This app is based on ASAB."

	def __init__(self, args=None):

		try:
			# EX_OK code is not available on Windows
			self.ExitCode = os.EX_OK
		except AttributeError:
			self.ExitCode = 0

		# Parse command line
		args = self.parse_arguments(args=args)

		# Load configuration
		Config._load()

		if hasattr(args, "daemonize") and args.daemonize:
			self.daemonize()

		elif hasattr(args, "kill") and args.kill:
			self.daemon_kill()

		# Seed the random generator
		random.seed()

		# Obtain the event loop
		self.Loop = asyncio.get_event_loop()
		if self.Loop.is_closed():
			self.Loop = asyncio.new_event_loop()
			asyncio.set_event_loop(self.Loop)

		self.LaunchTime = time.time()
		self.BaseTime = self.LaunchTime - self.Loop.time()

		# Setup logging
		self.Logging = Logging(self)

		# Configure the event loop
		self.Loop.set_exception_handler(_loop_exception_handler)
		if Config["logging"].getboolean("verbose"):
			self.Loop.set_debug(True)

		# Adding a handler to listen to the interrupt event
		if platform.system() == "Windows":

			# Windows win32api import
			import win32api

			def handler(type):
				self.stop()
				return True

			win32api.SetConsoleCtrlHandler(handler, True)

		else:

			# POSIX and other reasonable systems
			self.Loop.add_signal_handler(signal.SIGINT, self.stop)
			self.Loop.add_signal_handler(signal.SIGTERM, self.stop)
			self.Loop.add_signal_handler(signal.SIGHUP, self._hup)

		self._stop_event = asyncio.Event(loop=self.Loop)
		self._stop_event.clear()
		self._stop_counter = 0

		from .pubsub import PubSub
		self.PubSub = PubSub(self)

		self.Modules = []
		self.Services = {}

		# Setup ASAB API
		if len(Config['asab:web']["listen"]) > 0:
			from asab.api import Module
			self.add_module(Module)

		L.info("Initializing ...")


	def create_argument_parser(
		self,
		prog=None,
		usage=None,
		description=None,
		epilog=None,
		prefix_chars='-',
		fromfile_prefix_chars=None,
		argument_default=None,
		conflict_handler='error',
		add_help=True
	):
		'''
		This method can be overriden to adjust argparse configuration.
		Refer to the Python standard library to `argparse.ArgumentParser` for details of arguments.
		'''

		parser = argparse.ArgumentParser(
			prog=prog,
			usage=usage,
			description=description if description is not None else self.Description,
			epilog=epilog,
			formatter_class=argparse.RawDescriptionHelpFormatter,
			prefix_chars=prefix_chars,
			fromfile_prefix_chars=fromfile_prefix_chars,
			argument_default=argument_default,
			conflict_handler=conflict_handler,
			add_help=add_help
		)
		parser.add_argument('-c', '--config', help='specify a path to a configuration file')
		parser.add_argument('-v', '--verbose', action='store_true', help='print more information (enable debug output)')
		parser.add_argument('-s', '--syslog', action='store_true', help='enable logging to a syslog')
		parser.add_argument('-l', '--log-file', help='specify a path to a log file')
		parser.add_argument('-w', '--web-api', help='activate Asab web API (default listening port is 0.0.0.0:8080)', const="0.0.0.0:8080", nargs="?")


		if daemon is not None:
			parser.add_argument('-d', '--daemonize', action='store_true', help='run daemonized (in the background)')
			parser.add_argument('-k', '--kill', action='store_true', help='kill a running daemon and quit')

		return parser


	def parse_arguments(self, args=None):
		parser = self.create_argument_parser()
		args = parser.parse_args(args=args)

		if args.config is not None:
			Config._default_values['general']['config_file'] = args.config

		if args.verbose:
			Config._default_values['logging']['verbose'] = True

		if args.syslog:
			Config._default_values['logging:syslog']['enabled'] = True

		if args.log_file:
			Config._default_values['logging:file']['path'] = args.log_file

		if args.web_api:
				Config._default_values['asab:web']['listen'] = args.web_api
		return args


	def get_pidfile_path(self):
		pidfilepath = Config['general']['pidfile']
		if pidfilepath == "":
			return None
		elif pidfilepath == "!":
			return os.path.join('/var/run', os.path.basename(sys.argv[0]) + '.pid')
		else:
			return pidfilepath


	def daemonize(self):
		if daemon is None:
			print("Install 'python-daemon' module to support daemonizing.", file=sys.stderr)
			sys.exit(1)

		pidfilepath = self.get_pidfile_path()
		if pidfilepath is not None:
			pidfile = daemon.pidfile.TimeoutPIDLockFile(pidfilepath)


		working_dir = Config['general']['working_dir']

		uid = Config['general']['uid']
		if uid == "":
			uid = None

		gid = Config['general']['gid']
		if gid == "":
			gid = None

		signal_map = {
			signal.SIGTTIN: None,
			signal.SIGTTOU: None,
			signal.SIGTSTP: None,
		}

		self.DaemonContext = daemon.DaemonContext(
			working_directory=working_dir,
			signal_map=signal_map,
			pidfile=pidfile,
			uid=uid,
			gid=gid,
		)

		try:
			self.DaemonContext.open()
		except lockfile.AlreadyLocked as e:
			print("Cannot create a PID file '{}':".format(pidfilepath), e, file=sys.stderr)
			sys.exit(1)


	def daemon_kill(self):
		if daemon is None:
			print("Install 'python-daemon' module to support daemonising.", file=sys.stderr)
			sys.exit(1)

		pidfilepath = self.get_pidfile_path()
		if pidfilepath is None:
			sys.exit(0)

		try:
			pid = open(pidfilepath, "r").read()
		except FileNotFoundError:
			print("Pid file '{}' not found.".format(pidfilepath), file=sys.stderr)
			sys.exit(0)

		pid = int(pid)

		for sno in [signal.SIGINT, signal.SIGINT, signal.SIGINT, signal.SIGINT, signal.SIGTERM]:
			try:
				os.kill(pid, sno)
			except ProcessLookupError:
				print("Process with pid '{}' not found.".format(pid), file=sys.stderr)
				sys.exit(0)
			for i in range(10):
				if not os.path.exists(pidfilepath):
					sys.exit(0)
				time.sleep(0.1)
			print("Daemon process (pid: {}) still running ...".format(pid), file=sys.stderr)

		print("Pid file '{}' not found.".format(pidfilepath), file=sys.stderr)
		sys.exit(1)



	def run(self):
		# Comence init-time governor

		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.initialize(),
				self._init_time_governor(asyncio.Future()),
			],
			return_when=asyncio.FIRST_EXCEPTION
		))

		for task in finished_tasks:
			# This one also raises exceptions from futures, which is perfectly ok
			task.result()
		if len(pending_tasks) > 0:
			raise RuntimeError("Failed to fully initialize. Here are pending tasks: {}".format(pending_tasks))

		# Comence run-time and application main() function
		L.info("Running ...")
		self._stop_event.clear()
		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.main(),
				self._run_time_governor(asyncio.Future()),
			],
			return_when=asyncio.FIRST_EXCEPTION
		))
		for task in finished_tasks:
			try:
				task.result()
			except BaseException:
				L.exception("Exception in {}".format(task))

		# TODO: Process pending_tasks tasks from above

		# Comence exit-time
		L.info("Exiting ...")
		finished_tasks, pending_tasks = self.Loop.run_until_complete(asyncio.wait(
			[
				self.finalize(),
				self._exit_time_governor(asyncio.Future()),
			],
			return_when=asyncio.FIRST_EXCEPTION
		))
		for task in finished_tasks:
			try:
				task.result()
			except BaseException:
				L.exception("Exception in {}".format(task))

		# TODO: Process pending_tasks tasks from above (should be none)

		# Python 3.5 lacks support for shutdown_asyncgens()
		if hasattr(self.Loop, "shutdown_asyncgens"):
			self.Loop.run_until_complete(self.Loop.shutdown_asyncgens())
		self.Loop.close()

		return self.ExitCode


	def stop(self, exit_code: int = None):
		if exit_code is not None:
			self.set_exit_code(exit_code)

		self._stop_event.set()
		self._stop_counter += 1
		self.PubSub.publish("Application.stop!", self._stop_counter)
		if self._stop_counter >= 3:
			L.fatal("Emergency exit")
			try:
				# EX_SOFTWARE code is not available on Windows
				return os._exit(os.EX_SOFTWARE)
			except AttributeError:
				return os._exit(0)


	def _hup(self):
		self.Logging.rotate()
		self.PubSub.publish("Application.hup!")

	# Modules

	def add_module(self, module_class):
		""" Load a new module. """

		for module in self.Modules:
			if isinstance(module, module_class):
				# Already loaded and registered
				return

		module = module_class(self)
		self.Modules.append(module)

		asyncio.ensure_future(module.initialize(self), loop=self.Loop)

	# Services

	def get_service(self, service_name):
		""" Get a new service by its name. """

		try:
			return self.Services[service_name]
		except KeyError:
			pass

		L.error("Cannot find service '{}' - not registered?".format(service_name))
		raise KeyError("Cannot find service '{}'".format(service_name))


	def _register_service(self, service):
		""" Register a new service using its name. """

		if service.Name in self.Services:
			L.error("Service '{}' already registered (existing:{} new:{})".format(
				service.Name, self.Services[service.Name], service))
			raise RuntimeError("Service {} already registered".format(service.Name))

		self.Services[service.Name] = service

		asyncio.ensure_future(service.initialize(self), loop=self.Loop)

	# Lifecycle callback

	async def initialize(self):
		pass

	async def main(self):
		pass

	async def finalize(self):
		pass

	# Governors

	async def _init_time_governor(self, future):
		self.PubSub.publish("Application.init!")
		future.set_result("initialize")


	async def _run_time_governor(self, future):
		timeout = Config.getint('general', 'tick_period')
		try:
			self.PubSub.publish("Application.run!")

			# Wait for stop event & tick in meanwhile
			for cycle_no in itertools.count(1):
				try:
					await asyncio.wait_for(self._stop_event.wait(), timeout=timeout)
					break
				except asyncio.TimeoutError:
					self.PubSub.publish("Application.tick!")
					if (cycle_no % 10) == 0:
						self.PubSub.publish("Application.tick/10!")
					if (cycle_no % 60) == 0:
						# Rebase a Loop time
						self.BaseTime = time.time() - self.Loop.time()
						self.PubSub.publish("Application.tick/60!")
					if (cycle_no % 300) == 0:
						self.PubSub.publish("Application.tick/300!")
					if (cycle_no % 600) == 0:
						self.PubSub.publish("Application.tick/600!")
					if (cycle_no % 1800) == 0:
						self.PubSub.publish("Application.tick/1800!")
					if (cycle_no % 3600) == 0:
						self.PubSub.publish("Application.tick/3600!")
					if (cycle_no % 43200) == 0:
						self.PubSub.publish("Application.tick/43200!")
					if (cycle_no % 86400) == 0:
						self.PubSub.publish("Application.tick/86400!")
					continue

		finally:
			future.set_result("run")


	async def _exit_time_governor(self, future):
		self.PubSub.publish("Application.exit!")

		# Finalize services
		futures = []
		for service in self.Services.values():
			nf = asyncio.ensure_future(service.finalize(self), loop=self.Loop)
			futures.append(nf)
		if len(futures) > 0:
			await asyncio.wait(futures, return_when=asyncio.ALL_COMPLETED)
			# TODO: Handle expections (if needed) - probably only print them

		# Finalize modules
		futures = []
		for module in self.Modules:
			nf = asyncio.ensure_future(module.finalize(self), loop=self.Loop)
			futures.append(nf)
		if len(futures) > 0:
			await asyncio.wait(futures, return_when=asyncio.ALL_COMPLETED)
			# TODO: Handle expections (if needed) - probably only print them

		tasks_awaiting = 0
		for i in range(3):
			try:
				ts = asyncio.all_tasks(self.Loop)
			except AttributeError:
				# Compatibility for Python 3.6-
				ts = asyncio.Task.all_tasks(self.Loop)
			tasks_awaiting = 0
			for t in ts:
				if t.done():
					continue
				tasks_awaiting += 1
			if tasks_awaiting <= 2:
				# 2 is for _exit_time_governor and wait()
				break

			await asyncio.sleep(1)

		else:
			L.warning("Exiting but {} async task(s) are still waiting".format(tasks_awaiting))

		future.set_result("exit")


	def set_exit_code(self, exit_code: int, force: bool = False):
		if (self.ExitCode < exit_code) or force:
			L.debug("Exit code set to {}".format(exit_code))
			self.ExitCode = exit_code

	# Time

	def time(self):
		'''
		Return UTC unix timestamp using a loop time (a fast way how to get a wall clock time).
		'''
		return self.BaseTime + self.Loop.time()



from .abc.service import Service


class StreamSocketServerService(Service):

	'''
	Example of use:

	class ServiceMyProtocolServer(asab.StreamSocketServerService):

		async def initialize(self, app):
			host = asab.Config.get('http', 'host')
			port = asab.Config.getint('http', 'port')

			L.debug("Starting server on {} {} ...".format(host, port))
			await self.create_server(app, MyProtocol, [(host, port)])
	'''

	def __init__(self, app, service_name):
		super().__init__(app, service_name)
		self._servers = []
		app.PubSub.subscribe("Application.exit!", self._on_exit)


	async def create_server(self, app, protocol, addrs):
		# TODO: Parallelize this ...
		for addr in addrs:
			host, port = addr
			server = await app.Loop.create_server(protocol, host, port)
			self._servers.append(server)


	def _on_exit(self, message_type):
		for server in self._servers:
			server.close()


	async def finalize(self, app):
		# TODO: Parallelize this ...
		for server in self._servers:
			await server.wait_closed()



import asyncio


class Timer(object):
	'''
	T.__init__(app, handler, autorestart=False) -> Timer.

	The relative and optionally repeating timer for asyncio.

	This class is simple relative timer that generate an event after a given time, and optionally repeating in regular intervals after that.

	:param app: An ASAB application.
	:param handler: A coro or future that will be called when a timer triggers.
	:param boolean autorestart: If `True` then a timer will be automatically restarted after triggering.

	:ivar Handler: A coro or future that will be called when a timer triggers.
	:ivar Task: A future that represent the timer task.
	:ivar App: An ASAB app.
	:ivar boolean AutoRestart: If `True` then a timer will be automatically restarted after triggering.

	The timer object is initialized as stopped.

	*Note*: The implementation idea is borrowed from "`Python - Timer with asyncio/coroutine <https://stackoverflow.com/questions/45419723/python-timer-with-asyncio-coroutine>`_" question on StackOverflow.
	'''

	def __init__(self, app, handler, autorestart=False):
		self.App = app
		self.Handler = handler
		self.Task = None
		self.AutoRestart = autorestart

		app.PubSub.subscribe("Application.stop!", self._on_stop)


	def start(self, timeout):
		'''
		Start the timer.

		:param float/int timeout: A timer delay in seconds.
		'''
		if self.is_started():
			raise RuntimeError("Timer is already started")
		self.Task = asyncio.ensure_future(self._job(timeout), loop=self.App.Loop)


	def stop(self):
		'''
		Stop the timer.
		'''
		if self.Task is not None:
			self.Task.cancel()
			self.Task = None


	def restart(self, timeout):
		'''
		Restart the timer.

		:param float/int timeout: A timer delay in seconds.
		'''

		if self.is_started():
			self.stop()
		self.start(timeout)


	def is_started(self):
		'''
		T.is_started() -> boolean
		Return `True` is the timer is started otherwise returns `False`.
		'''
		return self.Task is not None


	async def _job(self, timeout):
		await asyncio.sleep(timeout)
		self.Task = None
		if self.AutoRestart:
			self.start(timeout)
		await self.Handler()


	def _on_stop(self, message_type, n):
		# This is to ensure timer stop on application exit
		self.stop()



import aiohttp
import logging
import re

#

L = logging.getLogger(__name__)

#


def oauthclient_middleware_factory(app, *args, oauth_client_service, **kwargs):
	"""
	Serves to connect with the user info endpoint of OAuth 2.0 server to obtain identity of the user
	associated with the provided access token.

	The expected format of Authorization header is:
	Authorization: Bearer <ACCESS_TOKEN>
	X-Authorization-OAuth-Server: <OAUTH-SERVER-ID>

	For more information about user info, visit:
	https://connect2id.com/products/server/docs/api/userinfo
	"""

	# Bearer token Regex is based on RFC 6750
	# The OAuth 2.0 Authorization Framework: Bearer Token Usage
	# Chapter 2.1. Authorization Request Header Field
	AuthorizationHeaderRg = re.compile(r"^\s*Bearer ([A-Za-z0-9\-\.\+_~/=]*)")

	@aiohttp.web.middleware
	async def oauthclient_middleware(request, handler):

		# Check that the identity is not already inserted
		assert not hasattr(request, "Identity")

		authorization = request.headers.get(aiohttp.hdrs.AUTHORIZATION, None)
		if authorization is None:
			return await handler(request)

		am = AuthorizationHeaderRg.match(authorization)
		if am is None:
			L.warn("Authorization header '{}' is not in format.".format(authorization))
			return await handler(request)
		access_token = am.group(1)

		method, oauth_user_info, expired_at = oauth_client_service.UserInfoCache.get(access_token, (None, None, None))
		if expired_at is not None and expired_at < app.Loop.time():
			# Cache entry is expired
			del oauth_client_service.UserInfoCache[access_token]
			method, oauth_user_info, expired_at = None, None, None

		if oauth_user_info is None:
			oauth_server_id = request.headers.get('X-Authorization-OAuth-Server', None)
			method = oauth_client_service.get_method(oauth_server_id)
			if method is None:
				L.warn("Method for OAuth server id '{}' was not found.".format(oauth_server_id))
				return await handler(request)

			userinfo_url = method.Config["userinfo_url"]
			headers = {
				"Authorization": authorization,
			}
			async with aiohttp.ClientSession() as session:
				async with session.get(userinfo_url, headers=headers) as resp:
					if resp.status == 200:
						oauth_user_info = await resp.json()

						expired_at = app.Loop.time() + oauth_client_service.UserInfoCacheLongevity
						oauth_client_service.UserInfoCache[access_token] = (method, oauth_user_info, expired_at)
					else:
						# "authn_required_handler" decorator will then return "HTTPUnauthorized" to the client,
						# because of missing identity in the request
						L.warn("Call to OAuth server '{}' failed with status code '{}'.".format(userinfo_url, resp.status))

		if oauth_user_info is not None:
			request.UserInfo = oauth_user_info
			request.Identity = method.extract_identity(oauth_user_info)
		else:
			oauth_client_service.UserInfoCache.pop(access_token, None)

		return await handler(request)

	return oauthclient_middleware



import os
import sys
import glob
import logging
import inspect
import platform
import configparser
from collections.abc import MutableMapping


L = logging.getLogger(__name__)


class ConfigParser(configparser.ConfigParser):

	_syslog_sockets = {
		'Darwin': '/var/run/syslog'
	}

	_syslog_format = {
		'Darwin': 'm'
	}


	_default_values = {

		'general': {
			'config_file': os.environ.get('ASAB_CONFIG', ''),
			'tick_period': 1,  # In seconds
			'var_dir': os.path.expanduser('~/.' + os.path.splitext(os.path.basename(sys.argv[0]))[0]),

			# Daemonization
			'pidfile': '!',  # '!' has a special meaning => it transforms into platform specific location of pid file
			'working_dir': '.',
			'uid': '',
			'gid': '',
		},

		"logging": {
			'verbose': os.environ.get('ASAB_VERBOSE', False),
			"app_name": os.path.basename(sys.argv[0]),
			"sd_id": "sd",  # Structured data id, see RFC5424
			"level": "NOTICE",
			"levels": "",
		},

		"logging:console": {
			"format": "%(asctime)s %(levelname)s %(name)s %(struct_data)s%(message)s",
			"datefmt": "%d-%b-%Y %H:%M:%S.%f",
		},

		"logging:syslog": {
			"enabled": "false",
			# TODO: "facility": 'local1',
			"address": _syslog_sockets.get(platform.system(), "/dev/log"),
			"format": _syslog_format.get(platform.system(), "3"),
		},

		"logging:file": {
			"path": "",
			"format": "%(asctime)s %(levelname)s %(name)s %(struct_data)s%(message)s",
			"datefmt": "%d-%b-%Y %H:%M:%S.%f",
			"backup_count": 3,
			"backup_max_bytes": 0,
			"rotate_every": "",
		},

		"asab:web": {
			"listen": "",
		},


		# "passwords" section serves to securely store passwords
		# in the configuration file; the passwords are not
		# shown in the default API
		#
		# Usage in the configuration file:
		#
		# [connection:KafkaConnection]
		# password=${passwords:kafka_password}
		#
		# [passwords]
		# kafka_password=<MY_SECRET_PASSWORD>
		"passwords": {
		}
	}


	def add_defaults(self, dictionary):
		""" Add defaults to a current configuration """

		for section, keys in dictionary.items():
			section = str(section)

			if section not in self._sections:
				try:
					self.add_section(section)
				except ValueError:
					if self._strict:
						raise

			for key, value in keys.items():

				key = self.optionxform(str(key))
				if key in self._sections[section]:
					# Value exists, no default needed
					continue

				if value is not None:
					value = str(value)

				if value is not None and "$" in value:
					self.set(section, key, os.path.expandvars(value))
				else:
					self.set(section, key, value)


	def _traverse_includes(self, includes, this_dir):
		""" Reads included config files. Supports nested including. """

		if '\n' in includes:
			sep = '\n'
		else:
			sep = os.pathsep

		for include_glob in includes.split(sep):
			include_glob = os.path.expandvars(include_glob.strip())
			if len(include_glob) == 0:
				continue

			for include in glob.glob(include_glob):
				include = os.path.abspath(include)

				if include in self._included:
					# Preventing infinite dependency looping
					L.warn("Config file '{}' can be included only once.".format(include))
					continue

				self._included.add(include)
				self.set('general', 'include', '')

				self._load_dir_stack.append(os.path.dirname(include))
				try:
					self.read(include)
				finally:
					self._load_dir_stack.pop()

				includes = self.get('general', 'include', fallback='')
				self._traverse_includes(includes, os.path.dirname(include_glob))

	def _load(self):
		""" This method should be called only once, any subsequent call will lead to undefined behaviour """

		self._load_dir_stack = []

		config_fname = ConfigParser._default_values['general']['config_file']
		if config_fname != '':
			if not os.path.isfile(config_fname):
				print("Config file '{}' not found".format(config_fname), file=sys.stderr)
				sys.exit(1)

			self._load_dir_stack.append(os.path.dirname(config_fname))
			try:
				self.read(config_fname)
			finally:
				self._load_dir_stack.pop()

		includes = self.get('general', 'include', fallback='')

		self._included = set()
		self._traverse_includes(includes, this_dir=os.path.dirname(config_fname))

		self.add_defaults(ConfigParser._default_values)

		del self._load_dir_stack


class _Interpolation(configparser.ExtendedInterpolation):
	"""Interpolation which expands environment variables in values."""

	def before_read(self, parser, section, option, value):
		# Expand environment variables
		if '$' in value:
			os.environ['THIS_DIR'] = os.path.abspath(parser._load_dir_stack[-1])
			value = os.path.expandvars(value)

		return super().before_read(parser, section, option, value)


Config = ConfigParser(interpolation=_Interpolation())


class Configurable(object):

	'''
	Usage:
	class ConfigurableObject(asab.Configurable):

		ConfigDefaults = {
			'foo': 'bar',
		}

		def __init__(self, config_section_name, config=None):
			super().__init__(config_section_name=config_section_name, config=config)

			config_foo = self.Config.get('foo')

	'''

	ConfigDefaults = {}


	def __init__(self, config_section_name, config=None):
		self.Config = ConfigObjectDict()

		for base_class in inspect.getmro(self.__class__):
			if not hasattr(base_class, 'ConfigDefaults'):
				continue
			if len(base_class.ConfigDefaults) == 0:
				continue

			# Merge config defaults of each base class in the 'inheritance' way
			for key, value in base_class.ConfigDefaults.items():

				if value is None:
					raise ValueError("None value not allowed in ConfigDefaults. Found in %s:%s " % (config_section_name, key))

				if key not in self.Config:
					self.Config[key] = value

		if Config.has_section(config_section_name):
			for key, value in Config.items(config_section_name):
				self.Config[key] = value

		if config is not None:
			self.Config.update(config)


# This is for backward compatibility
ConfigObject = Configurable


class ConfigObjectDict(MutableMapping):


	def __init__(self):
		self._data = {}


	def __getitem__(self, key):
		return self._data[key]


	def __setitem__(self, key, value):
		self._data[key] = value


	def __delitem__(self, key):
		del self._data[key]


	def __iter__(self):
		return iter(self._data)


	def __len__(self):
		return len(self._data)


	def getboolean(self, key):
		value = self._data[key]
		if isinstance(value, bool):
			return value
		if value.lower() not in configparser.ConfigParser.BOOLEAN_STATES:
			raise ValueError('Not a boolean: %s' % value)
		return configparser.ConfigParser.BOOLEAN_STATES[value.lower()]


	def __repr__(self):
		return "<%s %r>" % (self.__class__.__name__, self._data)



import logging
from .. import Config, Module

from .service import LogManIOService

#

L = logging.getLogger(__name__)

#

Config.add_defaults(
	{
		'logman.io': {
			'url': 'amqps://{username}:{password}@feed.logman.io:5477/{virtualhost}',
			'username': 'testuser',
			'password': 'password',
			'virtualhost': 'playground',
			'routing_key': '',
		}
	}
)


class Module(Module):

	def __init__(self, app):
		super().__init__(app)
		self.service = LogManIOService(app, "asab.LogManIOService")



import json
import logging
import uuid
import datetime
import aiohttp.web
import fastjsonschema

#

L = logging.getLogger(__name__)
Lex = logging.getLogger("asab.web")


#

class JSONDumper(object):

	def __init__(self, pretty):
		self.pretty = pretty

	def __call__(self, obj):
		if self.pretty:
			return json.dumps(obj, indent=4, default=self.default) + '\n'
		else:
			return json.dumps(obj, default=self.default)

	def default(self, o):
		m = getattr(o, 'rest_get', None)
		if m is not None:
			return m()

		if isinstance(o, datetime.datetime):
			return o.isoformat()

		elif isinstance(o, bytes):
			return o.hex()

		try:
			return json.JSONEncoder.default(self, o)
		except TypeError:
			return "{}".format(o)


def json_response(request, data, pretty=None, dumps=JSONDumper, **kwargs):
	'''
	argument `dumps` allows to specify custom JSON dumper for a serialization.
	The default is `JSONDumper` class.

	## Pretty Result
	When appending ?pretty to any request made, the JSON returned will be pretty formatted (use it for debugging only!).
	'''
	assert issubclass(dumps, JSONDumper)
	pretty = request.query.get('pretty', 'no').lower() in frozenset(['true', '1', 't', 'y', 'yes', '']) or pretty

	return aiohttp.web.json_response(
		data,
		dumps=dumps(pretty),
		**kwargs
	)


@aiohttp.web.middleware
async def JsonExceptionMiddleware(request, handler):
	'''
	Installation of the handler to a web service:

	websvc.WebApp.middlewares.append(asab.web.rest.JsonExceptionMiddleware)
	'''

	try:
		response = await handler(request)
		return response

	# HTTP errors to JSON
	except aiohttp.web.HTTPError as ex:
		respdict = {
			'status': ex.status,
			'message': ex.text[5:]
		}
		if ex.status >= 400:
			euuid = uuid.uuid4()
			struct_data = {'uuid': str(euuid)}
			respdict['uuid'] = str(euuid)
			struct_data.update(request.headers)
			struct_data['path'] = request.path
			struct_data['status'] = ex.status
			Lex.error(ex, struct_data=struct_data)

		ex.content_type = 'application/json'
		ex.text = json.dumps(respdict, indent=4)
		raise ex

	# KeyError translates to 404
	except KeyError as e:
		euuid = uuid.uuid4()
		Lex.warning("KeyError when handling web request", exc_info=e, struct_data={'uuid': str(euuid)})

		if len(e.args) > 1:
			message = e.args[0] % e.args[1:]
		elif e.args[0] is None:
			message = "KeyError"
		else:
			message = e.args[0]

		return aiohttp.web.Response(
			text=json.dumps({
				"status": 404,
				"message": message,
				"uuid": str(euuid),
			}, indent=4),
			status=404,
			content_type='application/json'
		)

	# Other errors to JSON
	except Exception as e:
		euuid = uuid.uuid4()
		Lex.exception("Exception when handling web request", exc_info=e, struct_data={'uuid': str(euuid)})
		return aiohttp.web.Response(
			text=json.dumps({
				"status": 500,
				"message": "Internal Server Error",
				"uuid": str(euuid),
			}, indent=4),
			status=500,
			content_type='application/json'
		)


def json_schema_handler(json_schema, *_args, **_kwargs):

	"""
	The json schema handler implements validation of JSON documents by JSON schema.
	The library of fastjsonschema implements JSON schema drafts 04, 06 and 07
	https://horejsek.github.io/python-fastjsonschema/

	Examples of use:

	@asab.web.rest.json_schema_handler({
	'type': 'object',
	'properties': {
		'key1': {'type': 'string'},
		'key2': {'type': 'number'},
	}})
	async def login(self, request, *, json_data):
		...

	or by specifying json as file

	@asab.web.rest.json_schema_handler('./data/sample_json_schema.json')
	async def login(self, request, *, json_data):
		...

	Works for `application/json`, `application/x-www-form-urlencoded` and `multipart/form-data` post requests
	"""

	def decorator(func):
		# Initializing fastjsonschema.compile method and generating
		# the validation function for validating JSON schema

		# JSON schema set as a dict
		if isinstance(json_schema, dict):
			validate = fastjsonschema.compile(json_schema)

		# JSON schema set in a file
		elif isinstance(json_schema, str):
			with open(json_schema) as f:
				schema = json.load(f)
				validate = fastjsonschema.compile(schema)
		else:
			raise ValueError(
				"JSON schema input must be type <class 'dict'> or type <class 'str'>, "
				"not type {}.".format(type(json_schema)))

		form_content_types = frozenset(['', 'application/x-www-form-urlencoded', 'multipart/form-data'])

		async def validator(*args, **kwargs):
			# Initializing fastjsonschema.compile method and generating
			# the validation function for validating JSON schema
			request = args[-1]
			if request.content_type == 'application/json':
				data = await request.json()
			elif request.content_type in form_content_types:
				multi_dict = await request.post()
				data = {k: v for k, v in multi_dict.items()}
			else:
				Lex.warning(f"Unsupported content-type {request.content_type} for request {request}")
				raise aiohttp.web.HTTPBadRequest(reason=f"Unsupported content-type {request.content_type}")
			# Checking the validation on JSON data set
			try:
				validate(data)
				kwargs['json_data'] = data
			except fastjsonschema.exceptions.JsonSchemaException as e:
				Lex.warning(f"Can not validate request {request}. Reason: {e}")
				raise aiohttp.web.HTTPBadRequest(reason=str(e))
			except Exception as e:
				Lex.error(f"Unknown validation error for request {request}. Reason: {e}")
				raise e

			return await func(*args, **kwargs)

		return validator

	return decorator



import abc
import uuid
import hashlib
import datetime


class UpsertorABC(abc.ABC):

	def __init__(self, storage, collection, obj_id, version=None):
		'''
		'''

		self.Storage = storage
		self.Collection = collection
		self.ObjId = obj_id

		self.Version = version

		now = datetime.datetime.utcnow()
		self.ModSet = {
			'_m': now,  # Set the modification timestamp
		}
		if version == 0:
			self.ModSet['_c'] = now  # Set the creation timestamp

		self.ModUnset = {}

		self.ModInc = {
			'_v': 1,  # Increment '_v' at every change
		}

		self.ModPush = {}
		self.ModPull = {}


	def get_id_name(self):
		return "_id"


	@classmethod
	def generate_id(cls):
		m = hashlib.sha256()
		m.update(uuid.uuid4().bytes)
		return m.digest()


	def set(self, objField, value):
		'''
		Scalar set
		'''
		self.ModSet[objField] = value


	def unset(self, obj_field):
		'''
		Scalar unset
		'''
		self.ModUnset[obj_field] = ""


	def increment(self, field_name, amount=1):
		'''
		Scalar increment
		'''
		self.ModInc[field_name] = amount


	def decrement(self, field_name, amount=1):
		'''
		Scalar decrement
		'''
		return self.increment(field_name, -amount)


	def push(self, field_name, value):
		'''
		Push an item into a list
		'''
		if self.ModPush.get(field_name) is None:
			self.ModPush[field_name] = []
		self.ModPush[field_name].append(value)


	def pull(self, field_name, value):
		'''
		Pull an item from a list
		'''
		if self.ModPull.get(field_name) is None:
			self.ModPull[field_name] = []
		self.ModPull[field_name].append(value)


	@abc.abstractmethod
	async def execute(self):
		pass



# This is the backend for ASAB Web UI Kit.
import os
import glob
import functools
import aiohttp.web


class StaticDirProvider(object):

	def __init__(self, webapp, root, path, index='index.html'):
		self.Path = path
		self.Index = index
		self.Root = root.rstrip('/')

		webapp.router.add_get(self.Root if len(self.Root) > 0 else '/', self._index)

		# Read all files from path (build from React) and map them statically
		for fname in glob.glob(os.path.join(self.Path, '*')):
			if os.path.isfile(fname):
				webapp.router.add_get(self.Root + '/' + os.path.basename(fname), functools.partial(self._file, fname))
			elif os.path.isdir(fname):
				webapp.router.add_static(self.Root + '/' + os.path.basename(fname), path=fname)


	async def _index(self, request):
		return aiohttp.web.FileResponse(os.path.join(self.Path, self.Index))


	async def _file(self, fname, request):
		return aiohttp.web.FileResponse(fname)



import abc


class Metric(abc.ABC):

	def __init__(self, name: str, tags: dict):
		assert(name is not None)
		assert(tags is not None)
		self.Name = name
		self.Tags = tags


	@abc.abstractmethod
	def flush(self) -> dict:
		pass

	def rest_get(self):
		return {
			'Name': self.Name,
			'Tags': self.Tags,
		}


class Gauge(Metric):


	def __init__(self, name: str, tags: dict, init_values=None):
		super().__init__(name=name, tags=tags)
		self.Init = init_values
		self.Values = self.Init.copy()


	def set(self, name, value):
		self.Values[name] = value


	def flush(self) -> dict:
		return self.Values.copy()

	def rest_get(self):
		rest = super().rest_get()
		rest['Values'] = self.Values
		return rest


class Counter(Metric):


	def __init__(self, name, tags, init_values=None, reset: bool = True):
		super().__init__(name=name, tags=tags)
		self.Init = init_values if init_values is not None else dict()
		self.Values = self.Init.copy()
		self.Reset = reset


	def add(self, name, value, init_value=None):
		"""
		Adds to the counter specified by `name` the `value`.
		:param name: name of the counter
		:param value: value to be added to the counter
		:param init_value: init value, when the counter `name` is not yet set up (f. e. by init_values in the constructor)
		If None, KeyError will be raised.
		:return:
		"""

		try:
			self.Values[name] += value
		except KeyError as e:
			if init_value is None:
				raise e
			self.Values[name] = init_value + value


	def sub(self, name, value, init_value=None):
		"""
		Subtracts to the counter specified by `name` the `value`.
		:param name: name of the counter
		:param value: value to be subtracted from the counter
		:param init_value: init value, when the counter `name` is not yet set up (f. e. by init_values in the constructor)
		If None, KeyError will be raised.
		:return:
		"""

		try:
			self.Values[name] -= value
		except KeyError as e:
			if init_value is None:
				raise e
			self.Values[name] = init_value - value


	def flush(self) -> dict:
		ret = self.Values
		if self.Reset:
			self.Values = self.Init.copy()
		return ret


	def rest_get(self):
		rest = super().rest_get()
		rest['Values'] = self.Values
		return rest


class DutyCycle(Metric):
	'''
	https://en.wikipedia.org/wiki/Duty_cycle

		now = self.Loop.time()
		d = now - self.LastReadyStateSwitch
		self.LastReadyStateSwitch = now
	'''


	def __init__(self, loop, name: str, tags: dict, init_values=None):
		super().__init__(name=name, tags=tags)
		self.Loop = loop

		now = self.Loop.time()
		self.Values = {k: (v, now, 0.0, 0.0) for k, v in init_values.items()}


	def set(self, name, on_off: bool):
		now = self.Loop.time()
		v = self.Values.get(name)
		if v is None:
			self.Values[name] = (on_off, now, 0.0, 0.0)
			return

		if v[0] == on_off:
			return  # No change

		d = now - v[1]
		off_cycle = v[2]
		on_cycle = v[3]
		if on_off:
			# From off to on
			off_cycle += d
		else:
			# From on to off
			on_cycle += d

		self.Values[name] = (on_off, now, off_cycle, on_cycle)


	def flush(self) -> dict:
		now = self.Loop.time()
		ret = {}
		new_values = {}
		for k, v in self.Values.items():
			d = now - v[1]
			off_cycle = v[2]
			on_cycle = v[3]
			if v[0]:
				on_cycle += d
			else:
				off_cycle += d

			full_cycle = on_cycle + off_cycle
			if full_cycle > 0.0:
				ret[k] = on_cycle / full_cycle

			new_values[k] = (v[0], now, 0.0, 0.0)

		self.Values = new_values
		return ret


	def rest_get(self):
		rest = super().rest_get()
		rest['Values'] = self.Values
		return rest



import abc
import logging
import asyncio
import asab

#

L = logging.getLogger(__name__)

#


class Broker(abc.ABC, asab.ConfigObject):
	"""
		Broker is implementation of object request broker (ORB) within the Message-oriented middleware concept.

		Broker allows to register callbacks for "task" and "reply" procedures, via "add" method.
		Tasks and replies are then distributed to the registered callbacks.

			self.Broker.add("task", self.task_handler)
			self.Broker.add("reply", self.reply_handler)

		In order to connect broker with the middleware (such as RabbitMQ, see AMQPBroker in the "amqp" submodule),
		it is needed that the broker is subscribed to task and reply queues in the middleware, via "subscribe" method.

			self.Broker.subscribe("task.queue")

		The method "publish" the serve to publish a task to the middleware.

			await self.Broker.publish("Hello world!", target="task")

		The brokers from different applications can be connected to the same middleware,
		where one application may publish tasks, while others process them and publish replies.
	"""


	def __init__(self, app, accept_replies: bool, task_service, config_section_name: str, config=None):
		if task_service is None:
			task_service = app.get_service("asab.MOMService")

		super().__init__(config_section_name=config_section_name, config=config)
		self.TaskService = task_service
		self.TaskService._register_broker(self)

		self.Loop = app.Loop
		self.Subscriptions = dict()
		self.Targets = {}
		self.AcceptReplies = accept_replies

		self.MainFuture = asyncio.ensure_future(self.main(), loop=self.Loop)


	async def finalize(self, app):
		self.MainFuture.cancel()


	def subscribe(self, subscription: str, **kwags):
		self.Subscriptions[subscription] = kwags
		asyncio.ensure_future(self.ensure_subscriptions(), loop=self.Loop)


	def add(self, target: str, handler):
		t = self.Targets.get(target)
		if t is None:
			self.Targets[target] = [handler]
		else:
			t.append(handler)


	async def dispatch(self, target, properties, body):
		tlist = self.Targets.get(target)
		if tlist is None:
			L.warning("Received a message for an unknown target '{}'".format(target))
			return

		for handler in tlist:
			reply = await handler(properties, body)
			if properties.reply_to is not None:
				# TODO: If reply_to is URL, then use HTTP to deliver reply
				await self.reply(
					reply,
					reply_to=properties.reply_to,
					correlation_id=properties.correlation_id,
				)
			elif reply is not None:
				L.warning("Discart the reply from target '{}'".format(target))



	async def main(self):
		pass


	async def ensure_subscriptions(self):
		pass


	async def publish(
		self,
		body,
		target: str = '',
		content_type: str = None,
		content_encoding: str = None,
		correlation_id: str = None,
		reply_to: str = None,
	):
		pass


	async def reply(
		self,
		body,
		reply_to: str,
		content_type: str = None,
		content_encoding: str = None,
		correlation_id: str = None,
	):
		pass



import aiohttp.web


def basicauth_middleware_factory(app, *args, **kwargs):

	@aiohttp.web.middleware
	async def basicauth_middleware(request, handler):
		auth_header = request.headers.get(aiohttp.hdrs.AUTHORIZATION)
		if auth_header is None:
			return await handler(request)

		auth = aiohttp.BasicAuth.decode(auth_header=auth_header, encoding='utf-8')
		request.Identity = auth.login

		# TODO: Check the password if required

		return await handler(request)

	return basicauth_middleware



import logging

from ..metrics.influxdb import influxdb_format


L = logging.getLogger(__name__)


class LogmanIOMetrics(object):


	def __init__(self, svc):
		self.Service = svc

	async def process(self, now, mlist):
		body = influxdb_format(now, mlist)
		await self.Service.OutboundQueue.put(('T', body))



import collections
import os
import shelve


class PersistentDict(dict):

	"""
	The persistent dictionary works as the regular Python dictionary but the content of the dictionary is stored in the file.
	You cat think of a ``PersistentDict`` as a simple `key-value store <https://en.wikipedia.org/wiki/Key-value_database>`_.
	It is not optimized for a frequent access. This class provides common ``dict`` interface.

	*Warning*: You must explicitly `load()` and `store()` content of the dictionary
	*Warning*: You can only store objects in the persistent dictionary that are serializable.
	"""

	def __init__(self, path):
		super().__init__()
		# Create directory, if needed
		dirname = os.path.dirname(path)
		if not os.path.isdir(dirname):
			os.makedirs(dirname)

		self._path = path

	def __delitem__(self, key):
		super().__delitem__(key)
		with shelve.open(self._path) as d:
			del d[key]

	def load(self) -> None:
		"""
		Load content of file as dictionary.
		"""
		with shelve.open(self._path) as d:
			for key, value in d.items():
				self[key] = value

	def store(self) -> None:
		"""
		Explicitly store content of persistent dictionary to file
		"""

		with shelve.open(self._path) as d:
			for key, value in self.items():
				d[key] = value

	def update(self, other=(), **kwds) -> None:
		"""
		Update D from mapping/iterable E and F.
		* If E present and has a .keys() method, does:     for k in E: D[k] = E[k]
		* If E present and lacks .keys() method, does:     for (k, v) in E: D[k] = v
		* In either case, this is followed by: for k, v in F.items(): D[k] = v

		Inspired by a https://github.com/python/cpython/blob/3.8/Lib/_collections_abc.py
		"""

		if isinstance(other, collections.Mapping):
			for key in other:
				self[key] = other[key]
		elif hasattr(other, "keys"):
			for key in other.keys():
				self[key] = other[key]
		else:
			for key, value in other:
				self[key] = value

		for key, value in kwds.items():
			self[key] = value



import asab
import asab.web

from .service import ApiService


class Module(asab.Module):

	def __init__(self, app):
		super().__init__(app)
		Module = asab.web.Module
		app.add_module(Module)

		self.service = ApiService(app, "asab.ApiService")



import asab

from .middleware import oauthclient_middleware_factory
from .method import ABCOAuthMethod
from .method import GitHubOAuthMethod
from .method import OpenIDConnectMethod

from .service import OAuthClientService


class Module(asab.Module):
	'''
	OAuth Client Module provides OAuthClientService and oauthclient_middleware_factory
	to connect with the user info endpoint of OAuth 2.0 server to obtain identity of the user
	associated with the provided access token.
	'''

	def __init__(self, app):
		super().__init__(app)
		self.Service = OAuthClientService(app, "asab.OAuthClientService")


__all__ = (
	'oauthclient_middleware_factory',
	'ABCOAuthMethod',
	'GitHubOAuthMethod',
	'OpenIDConnectMethod',
	'OAuthClientService',
)



import logging
import asyncio
import weakref
import functools


L = logging.getLogger(__name__)


class PubSub(object):


	def __init__(self, app):
		self.Subscribers = {}
		self.Loop = app.Loop


	def subscribe(self, message_type, callback):
		"""
		Subscribe a subscriber to the an message type.
		It could be even plain function, method or its coroutine variant (then it will be delivered in a dedicated future)
		"""

		# If subscribe is a bound method, do special treatment
		# https://stackoverflow.com/questions/53225/how-do-you-check-whether-a-python-method-is-bound-or-not
		if hasattr(callback, '__self__'):
			callback = weakref.WeakMethod(callback)

		else:
			callback = weakref.ref(callback)

		if message_type not in self.Subscribers:
			self.Subscribers[message_type] = [callback]
		else:
			self.Subscribers[message_type].append(callback)


	def subscribe_all(self, obj):
		"""
		Find all @asab.subscribe decorated methods on the obj and do subscription
		"""
		for member_name in dir(obj):
			member = getattr(obj, member_name)
			message_types = getattr(member, 'asab_pubsub_subscribe_to_message_types', None)
			if message_types is not None:
				for message_type in message_types:
					self.subscribe(message_type, member)


	def unsubscribe(self, message_type, callback):
		""" Remove a subscriber of an message type from the set. """

		callback_list = self.Subscribers.get(message_type)
		if callback_list is None:
			L.warning("Message type subscription '{}'' not found.".format(message_type))
			return

		remove_list = None

		for i in range(len(callback_list)):
			# Take an weakref entry in the callback list and references it
			c = callback_list[i]()

			# Check if a weak reference is working
			if c is None:  # a reference is lost, remove this entry
				if remove_list is None:
					remove_list = list()
				remove_list.append(callback_list[i])
				continue

			if c == callback:
				callback_list.pop(i)
				break

		else:
			L.warning("Subscriber '{}'' not found for the message type '{}'.".format(message_type, callback))

		if remove_list is not None:
			for callback_ref in remove_list:
				callback_list.remove(callback_ref)

		if len(callback_list) == 0:
			del self.Subscribers['message_type']


	def _callback_iter(self, message_type):

		def _deliver_async(loop, callback, message_type, *args, **kwargs):
			asyncio.ensure_future(callback(message_type, *args, **kwargs), loop=loop)

		callback_list = self.Subscribers.get(message_type)
		if callback_list is None:
			return

		remove_list = None

		for callback_ref in callback_list:
			callback = callback_ref()

			# Check if a weak reference is working
			if callback is None:  # a reference is lost
				if remove_list is None:
					remove_list = list()
				remove_list.append(callback_ref)
				continue

			if asyncio.iscoroutinefunction(callback):
				callback = functools.partial(_deliver_async, self.Loop, callback)

			yield callback

		if remove_list is not None:
			for callback_ref in remove_list:
				callback_list.remove(callback_ref)


	def publish(self, message_type, *args, **kwargs):
		""" Notify subscribers of an message type. Including arguments. """

		asynchronously = kwargs.pop('asynchronously', False)

		if asynchronously:
			for callback in self._callback_iter(message_type):
				self.Loop.call_soon(functools.partial(callback, message_type, *args, **kwargs))

		else:
			for callback in self._callback_iter(message_type):
				callback(message_type, *args, **kwargs)


###

class subscribe(object):

	'''
	Decorator

	Usage:

	@asab.subscribe("tick")
	def on_tick(self, message_type):
	print("Service tick")
	'''

	def __init__(self, message_type):
		self.message_type = message_type

	def __call__(self, f):
		if getattr(f, 'asab_pubsub_subscribe_to_message_types', None) is None:
			f.asab_pubsub_subscribe_to_message_types = [self.message_type]
		else:
			f.asab_pubsub_subscribe_to_message_types.append(self.message_type)
		return f


###

class Subscriber(object):

	'''
:any:`Subscriber` object allows to consume PubSub messages in coroutines.
It subscribes for various message types and consumes them.
It works on FIFO basis (First message In, first message Out).
If ``pubsub`` argument is None, the initial subscription is skipped.

.. code:: python

	subscriber = asab.Subscriber(
		app.PubSub,
		"Application.tick!",
		"Application.stop!"
	)
	'''

	def __init__(self, pubsub=None, *message_types):

		self._q = asyncio.Queue()
		self._subscriptions = []

		if pubsub is not None:
			for message_type in message_types:
				self.subscribe(pubsub, message_type)


	def subscribe(self, pubsub, message_type):
		'''
Subscribe for more message types. This method can be called many times with various ``pubsub`` objects.
		'''
		pubsub.subscribe(message_type, self)
		self._subscriptions.append((pubsub, message_type))


	def __call__(self, message_type, *args, **kwargs):
		self._q.put_nowait((message_type, args, kwargs))


	def message(self):
		'''
Wait for a message asynchronously.
Returns a three-members tuple ``(message_type, args, kwargs)``.

Example of the `await message()` use:

.. code:: python

	async def my_coroutine(app):
		# Subscribe for a two application events
		subscriber = asab.Subscriber(
			app.PubSub,
			"Application.tick!",
			"Application.exit!"
		)
		while True:
			message_type, args, kwargs = await subscriber.message()
			if message_type == "Application.exit!":
				break
			print("Tick.")

		'''
		return self._q.get()


	def __aiter__(self):
		return self


	async def __anext__(self):
		return await self._q.get()



import asab
from .forwarder import OAuthForwarder


class OAuthClientService(asab.Service):
	"""
	OAuthClientService serves to provide methods and cache to oauthclient_middleware_factory
	and register OAuthForwarder, which serves to add forward endpoints to the provided web server container,
	so the client applications may get/post information about the OAuth login.
	"""

	def __init__(self, app, service_name):
		super().__init__(app, service_name)
		self.App = app
		self.Methods = {}
		self.DefaultMethod = None
		self.Forwarder = None

		self.UserInfoCache = {}
		self.UserInfoCacheLongevity = 60  # In seconds

		app.PubSub.subscribe("Application.tick/60!", self.on_tick60)


	def append_method(self, method):
		self.Methods[method.Config["oauth_server_id"]] = method
		if self.DefaultMethod is None:
			self.DefaultMethod = method
		if self.Forwarder is not None:
			self.Forwarder.Methods = self.Methods


	def get_method(self, oauth_server_id):
		if oauth_server_id is None:
			return self.DefaultMethod
		return self.Methods[oauth_server_id]


	def configure(self, container, configure_forwarder=True, configure_middleware=True):
		"""
		Configure method configured OAuthForwarder and oauthclient_middleware to obtain information about OAuth
		identity and thus restrict access to the content.

		OAuthForwarderserves to add forward endpoints to the provided web server container,
		so the client applications may get/post information about the OAuth login.

		:param container: container which would be configured
		:param configure_forwarder: flag which specifies if to configure OAuthForwarder
		:param configure_middleware: flag which specifies if to configure oauthclient_middleware
		:return:
		"""

		if len(self.Methods) == 0:
			raise RuntimeError("OAuth methods need to be specified first. Call 'add_oauth_methods' before registering forwarder.")

		if configure_forwarder is True:
			self.Forwarder = OAuthForwarder(container=container, service=self)

		if configure_middleware is True:
			container.WebApp.middlewares.append(
				asab.web.authn.authn_middleware_factory(
					self.App,
					"oauth2client",
					oauth_client_service=self,
				)
			)


	def on_tick60(self, event_name):
		keys_to_remove = set()
		for key, (_, _, expired_at) in self.UserInfoCache.items():
			if expired_at < self.App.Loop.time():
				keys_to_remove.add(key)

		for key in keys_to_remove:
			del self.UserInfoCache[key]



import logging
import asab

from .service import MetricsService

#

L = logging.getLogger(__name__)

#

asab.Config.add_defaults(
	{
		'asab:metrics': {
			'target': '',  # Can be multiline
		}
	}
)


class Module(asab.Module):

	def __init__(self, app):
		super().__init__(app)
		self.service = MetricsService(app, "asab.MetricsService")



import aiohttp
import logging

import jwt
import base64
import json

from ...rest import json_response

#

L = logging.getLogger(__name__)

#


class OAuthForwarder(object):
	"""
	OAuthForwarder serves to add forward endpoints to the provided web server container,
	so the client applications may get/post information about the OAuth login.
	The forwarded endpoints include: POST token, GET identity, POST invalidate and POST forgot.

	OAuthForwarder expects specific instance of ABCOAuthMethod for every OAuth 2.0 server it forwards the requests to.
	Every OAuth 2.0 server is required to implement token and identity endpoints, while invalidate and forgot are optional.
	"""

	def __init__(self, *args, container, service, **kwargs):

		self.Service = service

		# Register endpoints in the provided container
		container.WebApp.router.add_post('/token', self.token)
		container.WebApp.router.add_get('/identity', self.identity)
		container.WebApp.router.add_post('/invalidate', self.invalidate)
		container.WebApp.router.add_post('/refresh', self.refresh)

	async def token(self, request):
		"""
		Forwards "Access Token Request", see https://tools.ietf.org/html/rfc6749#section-4.1.3
		:param request: received aiohttp request
		:return: json_response
		"""
		method = await self._get_method(request)
		if method is None:
			raise aiohttp.web.HTTPNotFound()
		if len(method.Config["token_url"]) == 0:
			raise aiohttp.web.HTTPNotFound()

		response = await self._forward_post(request, method.Config["token_url"], method.Config["client_id"], method.Config["client_secret"])

		# Extract the identity from the token right now and store it in the identity cache
		oauth_server_public_key = method.Config["oauth_server_public_key"].encode("utf-8")
		if self.Service.IdentityCache is not None and len(oauth_server_public_key) > 0 and response.get("status") == 200:
			content = response.get("content")
			if not isinstance(content, dict):
				L.warning("The received response '{}' was not in expected format.".format(json.dumps(response)))
				return json_response(request=request, data=response)

			# Receive access token and token ID from the response
			access_token = content.get("access_token")
			token_id = content.get("token_id")
			if access_token is None or token_id is None:
				L.warning("Skipping identity store. The received response '{}' was not in expected format.".format(json.dumps(response)))
				return json_response(request=request, data=response)

			# Decode token from token id: https://www.oauth.com/oauth2-servers/access-tokens/self-encoded-access-tokens/
			try:
				token = jwt.decode(token_id.encode("utf-8"), oauth_server_public_key, audience='client_id="{}"'.format(method.Config["client_id"]), algorithm='RS256')
			except (jwt.exceptions.DecodeError, jwt.exceptions.InvalidAudienceError):
				L.warning("Token ID '{}' could not be decoded. Did you specify the OAuth server '{}' public key correctly?".format(token_id, method.Config["oauth_server_id"]))
				return json_response(request=request, data=response)

			# Access token validation: https://openid.net/specs/openid-connect-core-1_0.html#3.1.3.8
			at_hash = token.get("at_hash")
			computed_at_hash = base64.urlsafe_b64encode(access_token[0:len(access_token) // 2].encode("utf-8")).decode("utf-8")
			if computed_at_hash != at_hash:
				L.warning("The access token '{}' from '{}' could not be validated.".format(access_token, method.Config["oauth_server_id"]))
				return json_response(request=request, data=response)

			# Identity format validation (OpenID Connect expects identity in "sub")
			identity = token.get("sub")
			if identity is None:
				L.warning("Identity from '{}' was not in expected format.".format(method.Config["oauth_server_id"]))
				return json_response(request=request, data=response)

			# Store the identity in cache
			oauth_server_id_access_token = "{}-{}".format(method.Config["oauth_server_id"], access_token)
			self.Service.IdentityCache[oauth_server_id_access_token] = ({"sub": identity}, identity)

		return json_response(request=request, data=response)

	async def identity(self, request):
		"""
		Forwards "UserInfo Request", see https://connect2id.com/products/server/docs/api/userinfo
		:param request: received aiohttp request
		:return: json_response
		"""
		method = await self._get_method(request)
		if method is None:
			raise aiohttp.web.HTTPNotFound()
		if len(method.Config["userinfo_url"]) == 0:
			raise aiohttp.web.HTTPNotFound()

		response = await self._forward_get(
			request,
			method.Config["userinfo"],
			method.Config["client_id"],
			method.Config["client_secret"]
		)
		return json_response(request=request, data=response)

	async def invalidate(self, request):
		"""
		Forwards "Revocation Request", see https://tools.ietf.org/html/rfc7009#page-4
		:param request: received aiohttp request
		:return: json_response
		"""
		method = await self._get_method(request)
		if method is None:
			raise aiohttp.web.HTTPNotFound()
		if len(method.Config["invalidate_url"]) == 0:
			raise aiohttp.web.HTTPNotFound()

		response = await self._forward_post(request, method.Config["invalidate_url"], method.Config["client_id"], method.Config["client_secret"])
		return json_response(request=request, data=response)

	async def refresh(self, request):
		"""
		Refreshes access token, see https://auth0.com/docs/tokens/refresh-token/current#use-a-refresh-token
		:param request: received aiohttp request
		:return: json_response
		"""
		method = await self._get_method(request)
		if method is None:
			raise aiohttp.web.HTTPNotFound()
		if len(method.Config["refresh_url"]) == 0:
			raise aiohttp.web.HTTPNotFound()

		response = await self._forward_post(request, method.Config["refresh_url"], method.Config["client_id"], method.Config["client_secret"])
		return json_response(request=request, data=response)

	async def _get_method(self, request):
		oauth_server_id = request.headers.get("X-OAuthServerId")

		if oauth_server_id is None:
			L.warning("The 'X-OAuthServerId' header was not provided.")
			return None

		method = self.Service.Methods.get(oauth_server_id)
		if method is None:
			L.warning("Method for OAuth server id '{}' was not found.".format(oauth_server_id))
			return None

		return method

	async def _forward_get(self, request, url, client_id, client_secret):
		# Construct headers
		headers = {}
		auth = request.headers.get("Authorization")
		if auth is not None:
			headers["Authorization"] = auth
		# Construct request
		data = dict(request.query)
		async with aiohttp.ClientSession() as session:
			async with session.get(url, headers=headers, params=data) as resp:
				response = {"status": resp.status}
				try:
					response["content"] = await resp.json()
					response["Content-Type"] = "application/json"
				except aiohttp.client_exceptions.ContentTypeError:
					L.warning("The response from '{}' with status code '{}' is not JSON. Using text instead.".format(url, resp.status))
					response["content"] = await resp.text()
					response["Content-Type"] = "text/html"
				return response

	async def _forward_post(self, request, url, client_id, client_secret):
		data = dict(await request.post())
		if len(client_id) != 0:
			data["client_id"] = client_id
		if len(client_secret) != 0:
			data["client_secret"] = client_secret
		async with aiohttp.ClientSession() as session:
			async with session.post(url, data=data) as resp:
				response = {"status": resp.status}
				try:
					response["content"] = await resp.json()
					response["Content-Type"] = "application/json"
				except aiohttp.client_exceptions.ContentTypeError:
					L.warning("The response from '{}' with status code '{}' is not JSON. Using text instead.".format(url, resp.status))
					response["content"] = await resp.text()
					response["Content-Type"] = "text/html"
				return response



import logging
import asab

#

L = logging.getLogger(__name__)

#


class Module(asab.Module):

	def __init__(self, app):
		super().__init__(app)
		sttype = asab.Config.get('asab:storage', 'type')

		if sttype == 'inmemory':
			from .inmemory import StorageService
			self.Service = StorageService(app, "asab.StorageService")

		elif sttype == 'mongodb':
			from .mongodb import StorageService
			self.Service = StorageService(app, "asab.StorageService")

		elif sttype == "elasticsearch":
			from .elasticsearch import StorageService
			self.Service = StorageService(app, "asab.StorageService")

		else:
			L.error("Unknown asab:storage type '{}'".format(sttype))



import itertools

import pika
from pkg_resources import parse_version


class QueueSubscriptionObject(object):

	def __init__(self, broker, queue_name, exchange: bool = False):
		self.Broker = broker
		self.QueueName = queue_name

		def on_channel_open(channel):
			channel.basic_qos(callback=on_qos_applied, prefetch_count=int(self.Broker.Config['prefetch_count']))

		def on_qos_applied(method):
			if parse_version(pika.__version__) > parse_version('1.0.a'):
				self.Channel.basic_consume(self.QueueName, on_consume_message)
			else:
				self.Channel.basic_consume(on_consume_message, self.QueueName)

		def on_consume_message(channel, method, properties, body):
			try:
				self.Broker.InboundQueue.put_nowait((channel, method, properties, body))
			except Exception:
				channel.basic_nack(method.delivery_tag, requeue=True)

		self.Channel = self.Broker.Connection.channel(on_open_callback=on_channel_open)


class ExchangeSubscriptionObject(object):
	'''
	This class handles a subscription to a topic exchange.
	It creates a temporary exclusive queue that is bound to a specified exchange.

	Usage:
	Broker.subscribe("amq.topic", topic="*.orange.*")

	... or for multiple topics via subscription to an exchange

	Broker.subscribe("amq.topic", exchange=True, routing_key=["*.orange.*", "*.*.rabbit"])
	'''

	QueueNumberSeq = itertools.count(1)

	def __init__(self, broker, exchange_name: str, exchange: bool = True, routing_key: str = None):
		self.Broker = broker
		self.QueueName = "~T{}@".format(next(self.QueueNumberSeq)) + broker.Origin
		self.ExchangeName = exchange_name
		if isinstance(routing_key, list):
			self.RoutingKey = routing_key
		else:
			self.RoutingKey = [routing_key]

		def on_channel_open(channel):
			channel.queue_declare(
				queue=self.QueueName,
				exclusive=True,
				auto_delete=True,
				callback=on_queue_declared,
			)

		if parse_version(pika.__version__) > parse_version('1.0.a'):
			def on_queue_declared(method):
				self.Channel.basic_qos(
					prefetch_count=int(self.Broker.Config['prefetch_count']),
					callback=on_qos_applied,
				)
				for rk in self.RoutingKey:
					self.Channel.queue_bind(
						queue=self.QueueName,
						exchange=self.ExchangeName,
						routing_key=rk
					)

			def on_qos_applied(method):
				self.Channel.basic_consume(self.QueueName, on_consume_message)

		else:  # pika < 1.0.0
			def on_queue_declared(method):
				self.Channel.basic_qos(
					callback=on_qos_applied,
					prefetch_count=int(self.Broker.Config['prefetch_count']),
				)
				for rk in self.RoutingKey:
					self.Channel.queue_bind(
						None,
						queue=self.QueueName,
						exchange=self.ExchangeName,
						routing_key=rk,
						nowait=True,  # We are not interested in the result
					)

			def on_qos_applied(method):
				self.Channel.basic_consume(on_consume_message, self.QueueName)

		def on_consume_message(channel, method, properties, body):
			try:
				self.Broker.InboundQueue.put_nowait((channel, method, properties, body))
			except Exception:
				channel.basic_nack(method.delivery_tag, requeue=True)

		self.Channel = self.Broker.Connection.channel(on_open_callback=on_channel_open)



import logging
import asyncio
import aiohttp

#

L = logging.getLogger(__name__)

#


class WebSocketFactory(object):

	'''
	websvc.WebApp.router.add_get('/api/ws', asab.web.WebSocketFactory(self))

	<script type="text/javascript">
		var connection = new WebSocket(((window.location.protocol === "https:") ? "wss://" : "ws://") + window.location.host + "/api/ws");
		connection.onmessage = function (e) {
			...
		};
	</script>

	'''


	def __init__(self, app, *, timeout=10.0, protocols=(), compress=True, max_msg_size=4194304):
		self.Loop = app.Loop
		self.WebSockets = set([])

		self.Timeout = timeout
		self.Protocols = protocols
		self.Compress = compress
		self.MaxMsgSize = max_msg_size

		app.PubSub.subscribe("Application.stop!", self._on_app_stop)


	def _on_app_stop(self, message_type, counter):
		# Clean up during application exit
		wslist = [ws.close(code=aiohttp.WSCloseCode.GOING_AWAY, message='Server shutdown') for ws in self.WebSockets]
		asyncio.gather(*wslist, loop=self.Loop)


	def send_parallely(self, send_futures):
		# THIS METHOD IS OBSOLETED, DON'T USE IT IN A NEW CODE.
		# Send messages in parallel
		asyncio.gather(*send_futures, loop=self.Loop)


	async def __call__(self, request):
		ws = await self.on_request(request)

		try:
			self.WebSockets.add(ws)

			async for msg in ws:
				await self.on_message(request, ws, msg)

		except asyncio.CancelledError:
			await ws.close()

		finally:
			await self.on_close(request, ws)
			self.WebSockets.remove(ws)

		return ws


	async def on_request(self, request):
		'''
		Override this to initialize websocket:

		async def on_request(self, request):
			...
			ws = super().on_request(request)
			...
			return ws

		'''
		ws = aiohttp.web.WebSocketResponse(
			timeout=self.Timeout,
			protocols=self.Protocols,
			compress=self.Compress,
			max_msg_size=self.MaxMsgSize,
		)
		session = request.get('Session')
		if session is not None:
			await session.Storage.set(session, ws)
		await ws.prepare(request)
		return ws


	async def on_message(self, request, websocket, message):
		'''
		Override this method to receive messages from client over the websocket
		'''
		pass


	async def on_close(self, request, websocket):
		'''
		Override this method to receive a notification that client closed the websocket connection
		'''
		pass



import motor.motor_asyncio
import pymongo
import bson

import asab
from .exceptions import DuplicateError
from .service import StorageServiceABC
from .upsertor import UpsertorABC

asab.Config.add_defaults(
	{
		'asab:storage': {
			'mongodb_uri': 'mongodb://localhost:27017',
			'mongodb_database': 'asabdb',
		}
	}
)


class StorageService(StorageServiceABC):
	'''
	Depends on `pymongo` and `motor`.
	'''


	def __init__(self, app, service_name, config_section_name='asab:storage'):
		super().__init__(app, service_name)
		self.Client = motor.motor_asyncio.AsyncIOMotorClient(asab.Config.get(config_section_name, 'mongodb_uri'))
		self.Database = self.Client[asab.Config.get(config_section_name, 'mongodb_database')]


	def upsertor(self, collection: str, obj_id=None, version=0):
		return MongoDBUpsertor(self, collection, obj_id, version)


	async def get(self, collection: str, obj_id) -> dict:
		coll = self.Database[collection]
		ret = await coll.find_one({'_id': obj_id})
		if ret is None:
			raise KeyError("NOT-FOUND")
		return ret


	async def get_by(self, collection: str, key: str, value) -> dict:
		"""
		Get object from collection by its key/value

		:param collection: Collection to get from
		:param key: Key to filter on
		:param value: Value to filter on
		:return: The object retrieved from a storage

		Raises:
			KeyError: If object{key: value} not found in `collection`
		"""
		coll = self.Database[collection]
		ret = await coll.find_one({key: value})
		if ret is None:
			raise KeyError("NOT-FOUND")
		return ret


	async def collection(self, collection: str) -> motor.motor_asyncio.AsyncIOMotorCollection:
		"""
		Get collection. Useful for custom operations

		:param collection: Collection to get
		:return: AsyncIOMotorCollection

		Examples:

			>>> coll = await storage.collection("test-collection")
			>>> cursor = coll.find({})
			>>> while await cursor.fetch_next:
			... 	obj = cursor.next_object()
			... 	pprint.pprint(obj)

		"""

		return self.Database[collection]


	async def delete(self, collection: str, obj_id):
		"""
		Delete object from `collection` by its `obj_id`

		:param collection: Collection to delete from
		:param obj_id: Object identification
		:return: `obj_id` -- Object identification

		Raises:
			KeyError: If `obj_id` not found in `collection`
		"""
		coll = self.Database[collection]
		ret = await coll.find_one_and_delete({'_id': obj_id})
		if ret is None:
			raise KeyError("NOT-FOUND")
		return ret['_id']


class MongoDBUpsertor(UpsertorABC):


	@classmethod
	def generate_id(cls):
		return bson.objectid.ObjectId()


	async def execute(self):
		id_name = self.get_id_name()
		addobj = {}

		if len(self.ModSet) > 0:
			addobj['$set'] = self.ModSet

		if len(self.ModInc) > 0:
			addobj['$inc'] = self.ModInc

		if len(self.ModPush) > 0:
			addobj['$push'] = {k: {'$each': v} for k, v in self.ModPush.items()}

		filter = {}

		if self.ObjId is not None:
			filter[id_name] = self.ObjId
		else:
			# We are going to insert a new object without explicit Id
			assert (self.Version == 0) or (self.Version is None)

		if self.Version is not None:
			filter['_v'] = int(self.Version)

		# First wave (adding stuff)
		if len(addobj) > 0:
			coll = self.Storage.Database[self.Collection]

			try:
				ret = await coll.find_one_and_update(
					filter,
					update=addobj,
					upsert=True if (self.Version == 0) or (self.Version is None) else False,
					return_document=pymongo.collection.ReturnDocument.AFTER
				)
			except pymongo.errors.DuplicateKeyError:
				assert(self.Version == 0)
				raise DuplicateError("Already exists", self.ObjId)

			self.ObjId = ret[id_name]

		# for k, v in self.ModUnset.items():
		# 	obj.pop(k, None)

		# for k, v in self.ModPull.items():
		# 	o = obj.pop(k, None)
		# 	if o is None: o = list()
		# 	for x in v:
		# 		try:
		# 			o.remove(x)
		# 		except ValueError:
		# 			pass
		# 	obj[k] = o

		return self.ObjId




class Singleton(type):

	"""
The `singleton pattern <https://en.wikipedia.org/wiki/Singleton_pattern>`_ is a software design pattern that restricts the instantiation of a class to one object.

*Note*: The implementation idea is borrowed from "`Creating a singleton in Python <https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python>`_" question on StackOverflow.
	"""

	_instances = {}

	def __call__(cls, *args, **kwargs):
		if cls not in cls._instances:
			cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
		return cls._instances[cls]

	@classmethod
	def delete(cls, singleton_cls):
		'''
		The method for an intentional removal of the singleton object.
		It shouldn't be used unless you really know what you are doing.

		One use case is a unit test, which removes an Application object after each iteration.
		'''
		del cls._instances[singleton_cls]



import abc


class Service(abc.ABC):
	"""
	Abstract service class
	"""

	def __init__(self, app, service_name):
		self.Name = service_name
		self.App = app
		app._register_service(self)

	# Lifecycle

	async def initialize(self, app):
		pass

	async def finalize(self, app):
		pass



import base64
import secrets

import cryptography.hazmat.primitives.ciphers
import cryptography.hazmat.primitives.ciphers.algorithms
import cryptography.hazmat.primitives.ciphers.modes
import cryptography.hazmat.backends

'''
This module provides AES GCM based payload protection.

Flow:

1. aes_gcm_generate_key() to get JWT AES 256 GCM key
2. Deliver the key to a Javascript app, also store AES key on the server
3. Import the key in the Javascript

	window.crypto.subtle.importKey(
		"jwk",
		jwt_aes_key,
		{ name: "AES-GCM" },
		false,
		["encrypt", "decrypt"]
	)
	.then(function(key){
		... use_the_key_here ...
	})
	.catch(function(err){
		console.error(err);
	});

4. Encrypt the payload in Javascript

	const iv = window.crypto.getRandomValues(new Uint8Array(12));
	window.crypto.subtle.encrypt(
		{
			name: "AES-GCM",
			iv: iv,
			// Optionally provide additionalData: ,
			tagLength: 128,
		},
		key, //from importKey above
		plaintext
	)
	.then(function(ciphertext){

		axios.post('/api/...',
			concatUint8ArrayAndArrayBuffer(iv, ciphertext),
		)
	}

5. Submit the payload to the server, don't send AES key with it!

5. Decrypt the payload on the server side
	plaintext = aes_gcm_decrypt(key = key_from_generate, ciphertext)

'''


def aes_gcm_decrypt(key: bytes, ciphertext: bytes, associated_data: bytes = None) -> bytes:
	'''
	Decrypt the ciphertext that is encrypted by AES GCM.
	'''

	iv = ciphertext[:12]
	message = ciphertext[12:-16]
	tag = ciphertext[-16:]

	# Construct a Cipher object, with the key, iv, and additionally the
	# GCM tag used for authenticating the message.
	decryptor = cryptography.hazmat.primitives.ciphers.Cipher(
		cryptography.hazmat.primitives.ciphers.algorithms.AES(key),
		cryptography.hazmat.primitives.ciphers.modes.GCM(iv, tag),
		backend=cryptography.hazmat.backends.default_backend()
	).decryptor()

	# We put associated_data back in or the tag will fail to verify
	# when we finalize the decryptor.
	if associated_data is not None:
		decryptor.authenticate_additional_data(associated_data)

	# Decryption gets us the authenticated plaintext.
	# If the tag does not match an InvalidTag exception will be raised.
	return decryptor.update(message) + decryptor.finalize()


def aes_gcm_generate_key(key: bytes = None) -> dict:
	'''
	Generate JWT AES 256 GCM key.
	'''

	if key is None:
		key = secrets.token_bytes(256 // 8)

	# JWT key
	return {
		"kty": "oct",
		"k": base64.urlsafe_b64encode(key).decode('ascii').rstrip('='),
		"alg": "A256GCM",
		"ext": True,
	}



import logging
import asab

from .service import ProactorService

#

L = logging.getLogger(__name__)

#

asab.Config.add_defaults(
	{
		'asab:proactor': {
			'max_workers': '0',
			'default_executor': True,
		}
	}
)


class Module(asab.Module):
	'''
	Proactor pattern based on loop.run_in_executor()

	https://en.wikipedia.org/wiki/Proactor_pattern
	'''

	def __init__(self, app):
		super().__init__(app)
		self.service = ProactorService(app, "asab.ProactorService")



import asyncio

from ..abc.service import Service
from ..config import Config

from .container import ZooKeeperContainer


class ZooKeeperService(Service):
	"""
	ZooKeeperService connects to Zookeeper via aiozk client:
	https://zookeeper.apache.org/
	https://pypi.org/project/aiozk/
	"""

	Config.add_defaults({
		"zookeeper": {  # create a default container, also ensures backward compatibility
			"urls": "",  # zookeeper:12181
			"path": "",  # /asab
		}
	})

	def __init__(self, app, service_name, config_section="zookeeper"):
		super().__init__(app, service_name)
		self.App = app
		self.DefaultContainer = None
		self.Containers = {}
		self.Futures = []

	async def initialize(self, app):
		# Create a default container
		# Default container ensures backward compatibility
		urls = Config["zookeeper"]["urls"]
		if len(urls) > 0:
			self.DefaultContainer = ZooKeeperContainer(app, "zookeeper")
			await self.DefaultContainer.initialize(app)

	async def finalize(self, app):
		if len(self.Futures) > 0:
			await asyncio.wait(self.Futures)
		if self.DefaultContainer is not None:
			await self.DefaultContainer.finalize(app)
		for containers in self.Containers.values():
			await containers.finalize(app)

	def register_container(self, container):
		self.Containers[container.ConfigSectionName] = container
		self.Futures.append(asyncio.ensure_future(container.initialize(self.App)))

	async def advertise(self, data, encoding="utf-8", container=None):
		if container is None:
			container = self.DefaultContainer
		if container is None:
			raise RuntimeError("The container must be specified.")
		return await container.advertise(data, encoding)

	async def get_children(self, container=None):
		if container is None:
			container = self.DefaultContainer
		if container is None:
			raise RuntimeError("The container must be specified.")
		return await container.get_children()

	async def get_data(self, child, encoding="utf-8", container=None):
		if container is None:
			container = self.DefaultContainer
		if container is None:
			raise RuntimeError("The container must be specified.")
		return await container.get_data(child, encoding)

	async def get_raw_data(self, child, container=None):
		if container is None:
			container = self.DefaultContainer
		if container is None:
			raise RuntimeError("The container must be specified.")
		return await container.get_raw_data(child)



import logging
import json
import asyncio

import aiohttp


L = logging.getLogger(__name__)


class LogManIOWebSocketUplink(object):


	def __init__(self, app, url, queue):
		self.App = app
		self.URL = url
		self.OutboundQueue = queue

		self.SenderFuture = None
		self.ReceiverFuture = None

		self.Running = True


	async def initialize(self, app):
		self._reconnect()


	async def finalize(self, app):
		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None

		if self.ReceiverFuture is not None:
			self.ReceiverFuture.cancel()
			self.ReceiverFuture = None


	def _reconnect(self):
		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None

		if self.ReceiverFuture is not None:
			self.ReceiverFuture.cancel()
			self.ReceiverFuture = None

		self.ReceiverFuture = asyncio.ensure_future(self._receiver(), loop=self.App.Loop)
		self.ReceiverFuture.add_done_callback(self._on_done)


	async def _receiver(self):
		try:
			async with aiohttp.ClientSession() as session:
				async with session.ws_connect(self.URL) as ws:
					await ws.ping()
					# We are connected :-)

					assert(self.SenderFuture is None)
					self.SenderFuture = asyncio.ensure_future(self._sender(ws), loop=self.App.Loop)
					self.SenderFuture.add_done_callback(self._on_done)

					async for msg in ws:
						pass

		except aiohttp.client_exceptions.ClientConnectorError as e:
			L.error("LogMan.io {}".format(e))


	async def _sender(self, ws):
		while self.Running:
			msg_type, body = await self.OutboundQueue.get()
			if isinstance(body, dict):
				body = json.dumps(body)
			await ws.send_str(msg_type + ' ' + body)


	def _on_done(self, future):
		try:
			future.result()
		except asyncio.CancelledError:
			pass
		except BaseException:
			L.exception("Error in LogMan.io websocket:")

		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None

		if self.ReceiverFuture is not None:
			self.ReceiverFuture.cancel()
			self.ReceiverFuture = None

			if self.Running:
				L.warning("LogMan.io reconnect in 5 secs.")
				self.App.Loop.call_later(5, self._reconnect)



from .middleware import authn_middleware_factory
from .middleware import authn_optional_handler
from .middleware import authn_required_handler
from .middleware import authorize_all
from .middleware import authorize_any

__all__ = (
	'authn_middleware_factory',
	'authn_optional_handler',
	'authn_required_handler',
	'authorize_all',
	'authorize_any',
)



import asab


class MOMService(asab.Service):
	"""
	MOMService implements message-oriented middleware:
	https://en.wikipedia.org/wiki/Message-oriented_middleware

	MOMService register brokers to post tasks to the task queue
	or process tasks and post them to the reply queue.
	"""


	def __init__(self, app, service_name):
		super().__init__(app, service_name)
		self.Brokers = set()


	def _register_broker(self, broker):
		self.Brokers.add(broker)


	async def finalize(self, app):
		for b in self.Brokers:
			await b.finalize(app)



import ssl
from ..config import ConfigObject


class SSLContextBuilder(ConfigObject):

	ConfigDefaults = {
		'cert': '',  # The certfile string must be the path to a PEM file containing the certificate as well as any number of CA certificates needed to establish the certificate’s authenticity.
		'key': '',  # The keyfile string, if present, must point to a file containing the private key in. Otherwise the private key will be taken from certfile as well.
		'password': '',
		'cafile': '',
		'capath': '',
		'ciphers': '',
		'dh_params': '',

		'verify_mode': '',  # empty or one of CERT_NONE, CERT_OPTIONAL or CERT_REQUIRED
		'check_hostname': '',
		'options': '',
	}

	def build(self, protocol=ssl.PROTOCOL_TLS):
		ctx = ssl.SSLContext(protocol=protocol)

		keyfile = self.Config.get("key")
		if len(keyfile) == 0:
			keyfile = None

		password = self.Config.get("password")
		if len(password) == 0:
			password = None

		cert = self.Config.get("cert")
		if len(cert) != 0:
			ctx.load_cert_chain(
				cert,
				keyfile=keyfile,
				password=password,
			)

		cafile = self.Config.get("cafile")
		if len(cafile) == 0:
			cafile = None

		capath = self.Config.get("capath")
		if len(capath) == 0:
			capath = None

		if (cafile is not None) or (capath is not None):
			ctx.load_verify_locations(cafile=cafile, capath=capath)

		ciphers = self.Config.get("ciphers")
		if len(ciphers) != 0:
			ctx.set_ciphers(ciphers)

		dh_params = self.Config.get("dh_params")
		if len(dh_params) != 0:
			ctx.load_dh_params(dh_params)

		verify_mode = self.Config.get("verify_mode")
		if len(verify_mode) > 0:
			verify_mode_tx = {
				'CERT_NONE': ssl.CERT_NONE,
				'CERT_OPTIONAL': ssl.CERT_OPTIONAL,
				'CERT_REQUIRED': ssl.CERT_REQUIRED,
			}.get(verify_mode.upper())
			if verify_mode_tx is None:
				raise RuntimeError("Unknown value {}".format(verify_mode))
			ctx.verify_mode = verify_mode_tx

		# TODO: check_hostname > ctx.check_hostname
		# TODO: options > ctx.options

		return ctx



import logging
import aiohttp
import urllib

import asab


L = logging.getLogger(__name__)


def influxdb_format(now, mlist):
	# CAREFUL: This function is used also in asab.logman.metrics
	rb = ""
	for metric, values in mlist:
		name = metric.Name

		field_set = []
		for fk, fv in values.items():
			if isinstance(fv, int):
				field_set.append("{}={}i".format(fk, fv))
			elif isinstance(fv, float):
				field_set.append("{}={}".format(fk, fv))
			elif isinstance(fv, str):
				field_set.append('{}="{}"'.format(fk, fv.replace('"', r'\"')))
			elif isinstance(fv, bool):
				field_set.append("{}={}".format(fk, 't' if fv else 'f'))
			else:
				raise RuntimeError("Unknown/invalud type of the metrics field: {} {}".format(type(fv), fk))

		for tk, tv in metric.Tags.items():
			name += ',{}={}'.format(tk, tv)

		rb += "{} {} {}\n".format(name, ','.join(field_set), int(now * 1e9))

	return rb


class MetricsInfluxDB(asab.ConfigObject):


	ConfigDefaults = {
		'url': 'http://localhost:8086/',
		'db': 'mydb',
		'username': '',
		'password': '',
	}


	def __init__(self, svc, config_section_name, config=None):
		super().__init__(config_section_name=config_section_name, config=config)

		self.WriteURL = '{}/write?db={}'.format(self.Config.get('url').rstrip('/'), self.Config.get('db'))

		username = self.Config.get('username')
		if username is not None and len(username) > 0:
			self.WriteURL += '&u={}'.format(urllib.parse.quote(username, safe=''))

		password = self.Config.get('password')
		if password is not None and len(password) > 0:
			self.WriteURL += '&p={}'.format(urllib.parse.quote(password, safe=''))

	async def process(self, now, mlist):

		rb = influxdb_format(now, mlist)

		async with aiohttp.ClientSession() as session:
			async with session.post(self.WriteURL, data=rb) as resp:
				response = await resp.text()
				if resp.status != 204:
					L.warning("Error when sending metrics to Influx: {}\n{}".format(resp.status, response))



import time
import hashlib
import collections


class Session(collections.MutableMapping):

	"""Session dict-like object."""

	def __init__(self, storage, id, new, max_age=None):
		self._id = id
		self._changed = False
		self._mapping = {}
		self._new = new
		self.MaxAge = max_age
		self.Storage = storage

		self._created = int(time.time())


	def __hash__(self):
		return hash((self.__class__, self._id))


	def __repr__(self):
		return '<{} [{}, new:{}, changed:{}, created:{} expired:{}] {!r}>'.format(
			self.__class__.__name__,
			hashlib.sha224(self._id.encode('utf-8')).hexdigest() if self._id is not None else "-",
			self._new, self._changed, self._created, self.is_expired(), self._mapping
		)


	def is_expired(self):
		expiration = self._created + self.MaxAge
		return time.time() >= expiration

	def reset(self):
		self._new = False
		self._change = False


	@property
	def Id(self):
		return self._id

	def set_id(self, id):
		if self._id is not None or not self._new:
			raise RuntimeError("Session id is already set!")
		self._id = id


	@property
	def Created(self):
		return self._created


	def is_new(self):
		return self._new

	def mark_changed(self):
		self._changed = True



	def __len__(self):
		return len(self._mapping)

	def __iter__(self):
		return iter(self._mapping)

	def __contains__(self, key):
		return key in self._mapping

	def __getitem__(self, key):
		return self._mapping[key]

	def __setitem__(self, key, value):
		self._mapping[key] = value
		self._changed = True

	def __delitem__(self, key):
		del self._mapping[key]
		self._changed = True



import abc

from ....config import ConfigObject


class ABCOAuthMethod(abc.ABC, ConfigObject):

	# Configuration for KeyCloak
	# "token_url": "http://localhost:8080/auth/realms/teskalabs/protocol/openid-connect/token",
	# "userinfo_url": "http://localhost:8080/auth/realms/teskalabs/protocol/openid-connect/userinfo",

	# Configuration for SeaCat Auth
	# "token_url": "http://localhost:8080/openidconnect/token",
	# "userinfo_url": "http://localhost:8080/openidconnect/userinfo",

	ConfigDefaults = {
		# TODO: Is "oauth_server_id" really needed?
		"oauth_server_id": "asab-oauth",

		# URL of "Access Token Request" endpoint, see https://tools.ietf.org/html/rfc6749#section-4.1.3
		"token_url": "http://localhost:8080/auth/realms/teskalabs/protocol/openid-connect/token",

		# URL of "UserInfo Request" endpoint, see https://connect2id.com/products/server/docs/api/userinfo
		"userinfo_url": "http://localhost:8080/auth/realms/teskalabs/protocol/openid-connect/userinfo",

		# URL of "Revocation Request" endpoint, see https://tools.ietf.org/html/rfc7009#page-4
		"invalidate_url": "",

		# Refreshes access token, see https://auth0.com/docs/tokens/refresh-token/current#use-a-refresh-token
		"refresh_url": "",

		"client_id": "",  # Client ID of the current application, see https://tools.ietf.org/html/rfc6749#section-4.1.1
		"client_secret": "",  # Client secret of the current application, see https://tools.ietf.org/html/rfc6749#section-4.1.1
		"oauth_server_public_key": "",  # To decode token from token id: https://www.oauth.com/oauth2-servers/access-tokens/self-encoded-access-tokens/
	}



	@abc.abstractmethod
	def extract_identity(self, oauth_user_info):
		pass


class GitHubOAuthMethod(ABCOAuthMethod):

	ConfigDefaults = {
		"oauth_server_id": "github.com",
		"token_url": "https://github.com/login/oauth/access_token",
		"userinfo_url": "https://api.github.com/user",
	}

	def __init__(self, config_section_name="oauth2:github", config=None):
		super().__init__(config_section_name=config_section_name, config=config)

	def extract_identity(self, oauth_user_info):
		return oauth_user_info["email"]


class OpenIDConnectMethod(ABCOAuthMethod):

	def __init__(self, config_section_name="oauth2:oidc", config=None):
		super().__init__(config_section_name=config_section_name, config=config)

	def extract_identity(self, oauth_user_info):
		return str(oauth_user_info["sub"])



import logging
import asyncio

from .. import Service, Config
from ..metrics.service import MetricsService

from .metrics import LogmanIOMetrics
from .log import LogmanIOLogHandler


L = logging.getLogger(__name__)


class LogManIOService(Service):

	def __init__(self, app, service_name):
		super().__init__(app, service_name)

		self.URL = Config.get('logman.io', 'url').format(
			username=Config.get('logman.io', 'username'),
			password=Config.get('logman.io', 'password'),
			virtualhost=Config.get('logman.io', 'virtualhost'),
		)

		self.OutboundQueue = asyncio.Queue(loop=app.Loop)

		from .amqp import LogManIOAMQPUplink as Uplink
		# from .websocket import LogManIOWebSocketUplink as Uplink
		self.Uplink = Uplink(app, self.URL, self.OutboundQueue)


	async def initialize(self, app):
		await super().initialize(app)
		await self.Uplink.initialize(app)


	async def finalize(self, app):
		await self.Uplink.finalize(app)
		await super().finalize(app)


	def configure_metrics(self, metrics_service):
		assert(isinstance(metrics_service, MetricsService))
		metrics_target = LogmanIOMetrics(self)
		metrics_service.add_target(metrics_target)


	def configure_logging(self, app):
		log_handler = LogmanIOLogHandler(self)
		app.Logging.RootLogger.addHandler(log_handler)



import asyncio
import logging
import os
import socket
import uuid

import pika
import pika.adapters.asyncio_connection

from .subscription import QueueSubscriptionObject, ExchangeSubscriptionObject
from ..broker import Broker

#

L = logging.getLogger(__name__)

#


class AMQPBroker(Broker):

	'''
The broker that uses Advanced Message Queuing Protocol (AMQP) and it can be used with e.g. RabbitMQ as a message queue.
	'''

	ConfigDefaults = {
		'url': 'amqp://username:password@localhost/virtualhost',
		'appname': 'asab.mom',
		'reconnect_delay': 10.0,
		'prefetch_count': 5,

		'exchange': 'amq.fanout',
		'reply_exchange': '',
	}

	def __init__(self, app, accept_replies=False, task_service=None, config_section_name="asab:mom:amqp", config=None):
		super().__init__(app, accept_replies, task_service, config_section_name, config)

		self.Origin = '{}#{}'.format(socket.gethostname(), os.getpid())

		self.Connection = None
		self.SubscriptionObjects = {}
		self.ReplyTo = None

		self.InboundQueue = asyncio.Queue(loop=app.Loop)
		self.OutboundQueue = asyncio.Queue(loop=app.Loop)

		self.SenderFuture = None

		self.Exchange = self.Config['exchange']
		self.ReplyExchange = self.Config['reply_exchange']


	async def finalize(self, app):
		await super().finalize(app)
		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None


	def _reconnect(self):
		if self.Connection is not None:
			if not (self.Connection.is_closing or self.Connection.is_closed):
				self.Connection.close()
			self.Connection = None

		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None

		parameters = pika.URLParameters(self.Config['url'])
		if parameters.client_properties is None:
			parameters.client_properties = dict()
		parameters.client_properties['application'] = self.Config['appname']

		self.SubscriptionObjects.clear()
		self.ReplyTo = None

		self.Connection = pika.adapters.asyncio_connection.AsyncioConnection(
			parameters=parameters,
			on_open_callback=self._on_connection_open,
			on_open_error_callback=self._on_connection_open_error,
			on_close_callback=self._on_connection_close
		)

	# Connection callbacks

	def _on_connection_open(self, connection):
		L.info("AMQP connected")
		asyncio.ensure_future(self.ensure_subscriptions(), loop=self.Loop)
		self.Connection.channel(on_open_callback=self._on_sending_channel_open)

	def _on_connection_close(self, connection, *args):
		try:
			code, reason = args
			L.warning("AMQP disconnected ({}): {}".format(code, reason))
		except ValueError:
			error, = args
			L.warning("AMQP disconnected: {}".format(error))
		self.Loop.call_later(float(self.Config['reconnect_delay']), self._reconnect)


	def _on_connection_open_error(self, connection, error_message=None):
		L.error("AMQP error: {}".format(error_message if error_message is not None else 'Generic error'))
		self.Loop.call_later(float(self.Config['reconnect_delay']), self._reconnect)


	def _on_sending_channel_open(self, channel):
		self.SenderFuture = asyncio.ensure_future(self._sender_future(channel), loop=self.Loop)


	async def ensure_subscriptions(self):
		if self.Connection is None:
			return
		if not self.Connection.is_open:
			return

		for s, pkwargs in self.Subscriptions.items():
			if s in self.SubscriptionObjects:
				continue
			if pkwargs.get('exchange', False):
				self.SubscriptionObjects[s] = ExchangeSubscriptionObject(self, s, **pkwargs)
			else:
				self.SubscriptionObjects[s] = QueueSubscriptionObject(self, s, **pkwargs)


	async def main(self):
		self._reconnect()
		while True:
			channel, method, properties, body = await self.InboundQueue.get()

			try:
				if self.AcceptReplies and (method.routing_key == self.ReplyTo):
					await self.dispatch("reply", properties, body)
				else:
					await self.dispatch(method.routing_key, properties, body)
			except BaseException:
				L.exception("Error when processing inbound message")
				channel.basic_nack(method.delivery_tag, requeue=False)
			else:
				channel.basic_ack(method.delivery_tag)


	async def publish(
		self,
		body,
		target: str = '',
		content_type: str = None,
		content_encoding: str = None,
		correlation_id: str = None,
		reply_to: str = None,
		exchange: str = None
	):
		await self.OutboundQueue.put((
			exchange if exchange is not None else self.Exchange,  # Where to publish
			target,  # Routing key
			body,
			pika.BasicProperties(
				content_type=content_type,
				content_encoding=content_encoding,
				delivery_mode=1,
				correlation_id=correlation_id,
				reply_to=self.ReplyTo,
				message_id=uuid.uuid4().urn,  # id
				app_id=self.Origin,  # origin
				# headers = { }
			)
		))


	async def reply(
		self,
		body,
		reply_to: str,
		content_type: str = None,
		content_encoding: str = None,
		correlation_id: str = None,
	):
		await self.OutboundQueue.put((
			self.ReplyExchange,  # Where to publish
			reply_to,  # Routing key
			body,
			pika.BasicProperties(
				content_type=content_type,
				content_encoding=content_encoding,
				delivery_mode=1,
				correlation_id=correlation_id,
				message_id=uuid.uuid4().urn,  # id
				app_id=self.Origin,  # origin
				# headers = { }
			)
		))


	async def _sender_future(self, channel):
		if self.AcceptReplies:
			self.ReplyTo = await self._create_exclusive_queue(channel, "~R@" + self.Origin)

		while True:
			exchange, routing_key, body, properties = await self.OutboundQueue.get()
			channel.basic_publish(exchange, routing_key, body, properties)


	async def _create_exclusive_queue(self, channel, queue_name):
		lock = asyncio.Event()
		lock.set()

		def on_queue_declared(method):
			lock.clear()
			assert(method.method.queue == queue_name)
			self.SubscriptionObjects[queue_name] = QueueSubscriptionObject(self, queue_name)

		channel.queue_declare(
			queue=queue_name,
			exclusive=True,
			auto_delete=True,
			callback=on_queue_declared,
		)

		await lock.wait()

		return queue_name



import logging
import re

import aiohttp

from .accesslog import AccessLogger
from ..config import ConfigObject
from ..net import SSLContextBuilder


class WebContainer(ConfigObject):

	'''
# Configuration examples

## Simple HTTP on 8080

[web]
listen=0.0.0.0 8080

## Multiple interfaces

[web]
listen:
	0.0.0.0 8080
	:: 8080


## Multiple interfaces, one with HTTPS

[web]
listen:
	0.0.0.0 8080
	:: 8080
	0.0.0.0 8443 ssl:web
	0.0.0.0:8001

[ssl:web]
cert=...
key=...
...


## Multiple interfaces, one with HTTPS (inline)

[web]
listen:
	0.0.0.0 8080
	:: 8080
	0.0.0.0 8443 ssl
	0.0.0.0:8001

# The SSL parameters are inside of the WebContainer section
cert=...
key=...

...



	'''


	ConfigDefaults = {
		'listen': '0.0.0.0 8080',  # Can be multiline
		'backlog': 128,
		'rootdir': '',
		'servertokens': 'full',  # Controls whether 'Server' response header field is included ('full') or faked 'prod' ()
		'cors': '',
	}


	def __init__(self, websvc, config_section_name, config=None):
		super().__init__(config_section_name=config_section_name, config=config)

		self.BackLog = int(self.Config.get("backlog"))
		self.CORS = self.Config.get("cors")

		servertokens = self.Config.get("servertokens")
		if servertokens == 'prod':
			# Because we cannot remove token completely
			self.ServerTokens = "asab"
		else:
			from .. import __version__
			self.ServerTokens = aiohttp.web_response.SERVER_SOFTWARE + " asab/" + __version__

		# Parse listen address(es), can be multiline configuration item
		ls = self.Config.get("listen")
		self._listen = []
		for line in ls.split('\n'):
			line = line.strip()
			if len(line) == 0:
				continue

			if ' ' in line:
				line = re.split(r"\s+", line)
			else:
				# This line allows the (obsolete) format of IPv4 with ':'
				# such as "0.0.0.0:8001"
				line = re.split(r"[:\s]", line, 1)

			addr = line.pop(0).strip()
			port = line.pop(0).strip()
			port = int(port)
			ssl_context = None

			for param in line:
				if param.startswith('ssl:'):
					# Dedicated section for SSL
					ssl_context = SSLContextBuilder(param).build()
					# SSL parameters are included in the current config section
				elif param.startswith('ssl'):
					ssl_context = SSLContextBuilder("<none>", config=self.Config).build()
				else:
					raise RuntimeError("Unknown asab:web listen parameter: '{}'".format(param))
			self._listen.append((addr, port, ssl_context))

		self.WebApp = aiohttp.web.Application(loop=websvc.App.Loop)
		self.WebApp.on_response_prepare.append(self._on_prepare_response)
		self.WebApp['app'] = websvc.App

		rootdir = self.Config.get("rootdir")
		if len(rootdir) > 0:
			from .staticdir import StaticDirProvider
			self.WebApp['rootdir'] = StaticDirProvider(self.WebApp, root='/', path=rootdir)

		self.WebAppRunner = aiohttp.web.AppRunner(
			self.WebApp,
			handle_signals=False,
			access_log=logging.getLogger(__name__[:__name__.rfind('.')] + '.al'),
			access_log_class=AccessLogger,
		)

		websvc._register_container(self, config_section_name)
		websvc.App.PubSub.subscribe("Application.run!", self.start_container)

	async def initialize(self, app):
		pass

	async def start_container(self, event_type):
		await self.WebAppRunner.setup()

		for addr, port, ssl_context in self._listen:
			site = aiohttp.web.TCPSite(
				self.WebAppRunner,
				host=addr, port=port, backlog=self.BackLog,
				ssl_context=ssl_context,
			)
			await site.start()


	async def finalize(self, app):
		await self.WebAppRunner.cleanup()


	async def _on_prepare_response(self, request, response):
		response.headers['Server'] = self.ServerTokens

		if self.CORS == "*":
			response.headers['Access-Control-Allow-Origin'] = "*"
			response.headers['Access-Control-Allow-Methods'] = "GET, POST, DELETE, PUT, PATCH, OPTIONS"


	def get_ports(self):
		ports = []
		for addr, port, ssl_context in self._listen:
			ports.append(port)
		return ports



import asyncio
import logging
import platform

import pika
import pika.adapters.asyncio_connection

from .. import Config


L = logging.getLogger(__name__)


class LogManIOAMQPUplink(object):


	def __init__(self, app, url, queue):
		self.App = app
		self.OutboundQueue = queue

		self.Parameters = pika.URLParameters(url)
		if self.Parameters.client_properties is None:
			self.Parameters.client_properties = dict()
		self.Parameters.client_properties['application'] = 'asab.logman'

		self.SenderFuture = None
		self.Connection = None

		self.RoutingKey = Config.get('logman.io', 'routing_key')
		self.Hostname = platform.node()


	async def initialize(self, app):
		self._reconnect()


	async def finalize(self, app):
		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None


	def _reconnect(self):
		if self.Connection is not None:
			if not (self.Connection.is_closing or self.Connection.is_closed):
				self.Connection.close()
			self.Connection = None

		if self.SenderFuture is not None:
			self.SenderFuture.cancel()
			self.SenderFuture = None

		self.Connection = pika.adapters.asyncio_connection.AsyncioConnection(
			parameters=self.Parameters,
			on_open_callback=self._on_connection_open,
			on_open_error_callback=self._on_connection_open_error,
			on_close_callback=self._on_connection_close
		)


	def _on_connection_open(self, connection):
		L.info("LogMan.io connected")

		def _on_sending_channel_open(channel):
			self.SenderFuture = asyncio.ensure_future(self._sender_future(channel), loop=self.App.Loop)

		self.Connection.channel(on_open_callback=_on_sending_channel_open)


	def _on_connection_close(self, connection, *args):
		try:
			code, reason = args
			L.warning("LogMan.io disconnected ({}): {}".format(code, reason))
		except ValueError:
			error, = args
			L.warning("LogMan.io disconnected: {}".format(error))
		self.App.Loop.call_later(30, self._reconnect)


	def _on_connection_open_error(self, connection, error_message=None):
		L.error("LogMan.io error: {}".format(error_message if error_message is not None else 'Generic error'))
		self.App.Loop.call_later(30.0, self._reconnect)


	async def _sender_future(self, channel):
		while True:
			msg_type, body = await self.OutboundQueue.get()
			properties = pika.BasicProperties(
				content_type='application/json' if msg_type == 'sj' else 'text/plain',
				delivery_mode=2,  # Persistent delivery mode
				headers={
					'H': self.Hostname,
					'T': msg_type,
				}
			)

			channel.basic_publish('i', self.RoutingKey, body, properties)



import pathlib
import re
import subprocess

from setuptools.command.build_py import build_py
from setuptools import find_packages
from setuptools import setup

here = pathlib.Path(__file__).parent
txt = (here / 'asab' / '__init__.py').read_text('utf-8')
if (here / '.git').exists():
	# This branch is happening during build from git version
	module_dir = (here / 'asab')

	version = subprocess.check_output(
		['git', 'describe', '--abbrev=7', '--tags', '--dirty=+dirty', '--always'], cwd=module_dir)
	version = version.decode('utf-8').strip()
	if version[:1] == 'v':
		version = version[1:]

	build = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=module_dir)
	build = build.decode('utf-8').strip()

else:
	# This is executed from packaged & distributed version
	txt = (here / 'asab' / '__version__.py').read_text('utf-8')
	version = re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
	build = re.findall(r"^__build__ = '([^']+)'\r?$", txt, re.M)[0]


class custom_build_py(build_py):

	def run(self):
		super().run()

		# This replace content of `__version__.py` in build folder
		version_file_name = pathlib.Path(self.build_lib, 'asab/__version__.py')
		with open(version_file_name, 'w') as f:
			f.write("__version__ = '{}'\n".format(version))
			f.write("__build__ = '{}'\n".format(build))
			f.write("\n")


setup(
	name='asab',
	version=version,
	description='ASAB simplifies a development of async application servers',
	long_description=open('README.rst').read(),
	url='https://github.com/TeskaLabs/asab',
	author='TeskaLabs Ltd',
	author_email='info@teskalabs.com',
	license='BSD License',
	platforms='any',
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
	],
	keywords='asyncio',
	packages=find_packages(exclude=['module_sample']),
	project_urls={
		'Source': 'https://github.com/TeskaLabs/asab'
	},
	install_requires=[
		'aiohttp>=3.6.2',
		'fastjsonschema>=2.14.4',
		'aiozk>=0.25.0',
	],
	cmdclass={
		'build_py': custom_build_py,
	},
)



import aiohttp.web

import asab
import asab.web.rest


def tenant_middleware_factory(app, svc):
	"""
	Ensures the tenant is obtained from the TenantService.
	:param app: application object
	:param svc: TenantService
	:return: handler(request)
	"""

	@aiohttp.web.middleware
	async def tenant_middleware(request, handler):
		tenant_id = request.match_info.get('tenant')
		if tenant_id is not None:
			tenant = svc.locate_tenant(tenant_id)
			if tenant is not None:
				request.Tenant = tenant
		return await handler(request)

	return tenant_middleware


def tenant_handler(func):

	async def wrapper(*args, **kargs):
		request = args[-1]
		try:
			kargs['tenant'] = request.Tenant
		except AttributeError:
			return asab.web.rest.json_response(request, {
				'result': 'TENANT-NOT-FOUND',
			}, status=404)
		return await func(*args, **kargs)

	return wrapper



from .. import rest


class TenantWebHandler(object):

	def __init__(self, app, svc, web_container):

		self.TenantService = svc

		web_app = web_container.WebApp

		web_app.router.add_get('/tenants', self.get_tenants)

	async def get_tenants(self, request):

		tenants = self.TenantService.get_tenants()

		return rest.json_response(request, data=tenants)



from .service import StorageServiceABC
from .upsertor import UpsertorABC
from .exceptions import DuplicateError


class InMemoryUpsertor(UpsertorABC):


	async def execute(self):
		id_name = self.get_id_name()

		# Get the object
		if self.Version == 0:
			obj = {
				id_name: self.ObjId
			}
			self.Storage._set(self.Collection, self.ObjId, obj)

		else:
			obj = await self.Storage.get(self.Collection, self.ObjId)
			if obj is None:
				if self.Version is None:
					obj = {
						id_name: self.ObjId
					}
					self.Storage._set(self.Collection, self.ObjId, obj)
				else:
					raise RuntimeError("Previous version of '{}' not found".format(self.ObjId))


		for k, v in self.ModSet.items():
			obj[k] = v

		for k, v in self.ModUnset.items():
			obj.pop(k, None)

		for k, v in self.ModInc.items():
			o = obj.pop(k, 0)
			obj[k] = o + v

		for k, v in self.ModPush.items():
			o = obj.pop(k, None)
			if o is None:
				o = list()
			o.extend(v)
			obj[k] = o

		for k, v in self.ModPull.items():
			o = obj.pop(k, None)
			if o is None:
				o = list()
			for x in v:
				try:
					o.remove(x)
				except ValueError:
					pass
			obj[k] = o

		return self.ObjId


class StorageService(StorageServiceABC):


	def __init__(self, app, service_name):
		super().__init__(app, service_name)
		self.InMemoryCollections = {}


	def upsertor(self, collection: str, obj_id=None, version=0):
		return InMemoryUpsertor(self, collection, obj_id, version)


	async def get(self, collection: str, obj_id):
		coll = self.InMemoryCollections[collection]
		return coll[obj_id]


	async def get_by(self, collection: str, key: str, value):
		"""
		Raises:
			NotImplementedError: Not implemented on InMemoryStorage
		"""
		raise NotImplementedError()


	async def delete(self, collection: str, obj_id):
		"""
		Delete object from `collection` by its `obj_id`

		:param collection: Collection to delete from
		:param obj_id: Object identification

		Raises:
			KeyError: If `obj_id` not found in `collection`
		"""
		coll = self.InMemoryCollections[collection]
		del coll[obj_id]


	def _set(self, collection: str, obj_id, obj):
		try:
			coll = self.InMemoryCollections[collection]
		except KeyError:
			coll = {}
			self.InMemoryCollections[collection] = coll

		nobj = coll.setdefault(obj_id, obj)
		if nobj != obj:
			raise DuplicateError("Already exists", obj_id)



import abc


class SessionStorage(abc.ABC):


	def __init__(self, app, max_age, session_class):
		self.App = app
		self.MaxAge = max_age
		self.SessionClass = session_class


	@abc.abstractmethod
	async def load(self, request):
		'''
		Load a session object from a persistent store
		or create a new one using create() method, if needed
		'''
		raise NotImplementedError()


	@abc.abstractmethod
	def set(self, session, response):
		'''
		Set session object reference to a response
		'''
		raise NotImplementedError()


	@abc.abstractmethod
	async def delete(self, session):
		'''
		Delete session object from a persistent store
		'''
		raise NotImplementedError()


	@abc.abstractmethod
	async def store(self, session):
		'''
		Store session object to persistent store (if needed)
		'''
		raise NotImplementedError()


	async def create(self, request):
		return self.SessionClass(self, id=None, new=True, max_age=self.MaxAge)



import logging

from .service import ZooKeeperService
from .container import ZooKeeperContainer

from ..abc.module import Module

#

L = logging.getLogger(__name__)

#


class Module(Module):

	def __init__(self, app):
		super().__init__(app)
		self.Service = ZooKeeperService(app, "asab.ZooKeeperService")



import logging
import datetime
import os
import sys
import platform
import json

from ..log import LOG_NOTICE


class LogmanIOLogHandler(logging.Handler):

	def __init__(self, svc, level=logging.NOTSET):
		super().__init__(level=level)

		self.Service = svc

		self.Facility = None  # TODO: Read this from config
		self.Pid = os.getpid()
		self.Environment = None
		self.Hostname = platform.node()
		self.Program = os.path.basename(sys.argv[0])


	def emit(self, record):
		if record.name == 'asab.metrics.service' and record.levelno == LOG_NOTICE:
			return  # No metrics to be submitted this way

		if record.levelno > logging.DEBUG and record.levelno <= logging.INFO:
			severity = 6  # Informational
		elif record.levelno <= LOG_NOTICE:
			severity = 5  # Notice
		elif record.levelno <= logging.WARNING:
			severity = 4  # Warning
		elif record.levelno <= logging.ERROR:
			severity = 3  # Error
		elif record.levelno <= logging.CRITICAL:
			severity = 2  # Critical
		else:
			severity = 1  # Alert

		log_entry = {
			"@timestamp": datetime.datetime.utcfromtimestamp(record.created).isoformat() + 'Z',
			"T": "syslog",
			"H": self.Hostname,
			"P": self.Program,
			"C": record.name,
			"s": "{}:{}".format(record.funcName, record.lineno),
			"p": record.process,
			"Th": record.thread,
			"l": severity,
		}

		message = record.getMessage()
		if record.exc_text is not None:
			message += '\n' + record.exc_text
		if record.stack_info is not None:
			message += '\n' + record.stack_info
		if len(message) > 0:
			log_entry['M'] = message

		if self.Facility is not None:
			log_entry['f'] = self.Facility

		if self.Environment is not None:
			log_entry['e'] = self.Environment

		sd = record.__dict__.get("_struct_data")
		if sd is not None:
			log_entry['sd'] = sd

		self.Service.OutboundQueue.put_nowait(('sj', json.dumps(log_entry)))



import aiohttp
import logging
import datetime
import json

import asab
from .service import StorageServiceABC
from .upsertor import UpsertorABC

#
L = logging.getLogger(__name__)
#

asab.Config.add_defaults(
	{
		'asab:storage': {
			'elasticsearch_url': 'http://localhost:9200/',
			'elasticsearch_username': '',
			'elasticsearch_password': '',
		}
	}
)


class StorageService(StorageServiceABC):
	"""
	Depends on `aiohttp`.
	"""

	def __init__(self, app, service_name, config_section_name='asab:storage'):
		super().__init__(app, service_name)
		self.Loop = app.Loop

		self.ESURL = asab.Config.get(config_section_name, 'elasticsearch_url')
		# self._timeout = asab.Config.get(config_section_name, 'elasticsearch_timeout')

		username = asab.Config.get(config_section_name, 'elasticsearch_username')
		password = asab.Config.get(config_section_name, 'elasticsearch_password')

		if username == '':
			self._auth = None
		else:
			self._auth = aiohttp.BasicAuth(login=username, password=password)

		self._ClientSession = None

	async def finalize(self, app):
		if self._ClientSession is not None and not self._ClientSession.closed:
			await self._ClientSession.close()
			self._ClientSession = None

	def session(self):
		if self._ClientSession is None:
			self._ClientSession = aiohttp.ClientSession(auth=self._auth, loop=self.Loop)
		elif self._ClientSession.closed:
			self._ClientSession = aiohttp.ClientSession(auth=self._auth, loop=self.Loop)
		return self._ClientSession

	async def delete(self, index, _id=None):
		if _id:
			url = "{}{}/_doc/{}".format(self.ESURL, index, _id)
		else:
			url = "{}{}".format(self.ESURL, index)
		async with self.session().request(method="DELETE", url=url) as resp:
			assert resp.status == 200, "Unexpected response code: {}".format(resp.status)
			resp = await resp.json()
			if resp.get("acknowledged", False):
				return resp
			assert resp["result"] == "deleted", "Document was not deleted"
			return resp

	async def get_by(self, collection: str, key: str, value):
		raise NotImplementedError("get_by")

	async def get(self, index: str, obj_id) -> dict:
		url = "{}{}/_doc/{}".format(self.ESURL, index, obj_id)
		async with self.session().request(method="GET", url=url) as resp:
			obj = await resp.json()

		# Manipulate the output so that it is a requested object
		ret = obj['_source']
		ret['_v'] = obj['_version']
		ret['_id'] = obj['_id']

		return ret

	def upsertor(self, index: str, obj_id=None, version: int = 0):
		return ElasicSearchUpsertor(self, index, obj_id, version)

	async def list(self, index, size=10000):
		'''
		Custom ElasticSearch method
		'''
		url = "{}{}/_search?size={}".format(self.ESURL, index, size)
		async with self.session().request(method="GET", url=url) as resp:
			assert resp.status == 200, "Unexpected response code: {}".format(resp.status)
			return await resp.json()

	async def indices(self, search_string=None):
		'''
		Custom ElasticSearch method
		'''
		url = "{}_cat/indices/{}?format=json".format(self.ESURL, search_string)
		async with self.session().request(method="GET", url=url) as resp:
			assert resp.status == 200, "Unexpected response code: {}".format(resp.status)
			return await resp.json()

	async def empty_index(self, index):
		'''
		Custom ElasticSearch method
		'''
		# TODO: There is an option here to specify settings (e.g. shard number, replica number etc) and mappings here
		url = "{}{}".format(self.ESURL, index)
		async with self.session().request(method="PUT", url=url) as resp:
			assert resp.status == 200, "Unexpected response code: {}".format(resp.status)
			return await resp.json()


class ElasicSearchUpsertor(UpsertorABC):

	@classmethod
	def generate_id(cls):
		raise NotImplementedError("generate_id")

	async def execute(self):
		if self.ObjId is None:
			return await self._insert_noobjid()
		else:
			return await self._upsert()

	async def _insert_noobjid(self):
		setobj = {}

		if len(self.ModSet) > 0:
			for k, v in self.ModSet.items():
				setobj[k] = serialize(self.ModSet[k])

		if len(self.ModInc) > 0:
			# addobj['$inc'] = self.ModInc
			# raise NotImplementedError("yet")
			pass

		if len(self.ModPush) > 0:
			# addobj['$push'] = {k: {'$each': v} for k, v in self.ModPush.items()}
			raise NotImplementedError("yet")

		# This is insert of the new document, the ObjId is to be generated by the ElasicSearch
		url = "{}{}/_doc".format(self.Storage.ESURL, self.Collection)
		async with self.Storage.session().request(method="POST", url=url, json=setobj) as resp:
			assert resp.status == 201, "Unexpected response code: {}".format(resp.status)
			resp_json = await resp.json()

		self.ObjId = resp_json['_id']

		return self.ObjId

	async def _upsert(self):
		upsertobj = {"doc": {}, "doc_as_upsert": True}

		if len(self.ModSet) > 0:
			for k, v in self.ModSet.items():
				upsertobj["doc"][k] = serialize(self.ModSet[k])
		url = "{}{}/_update/{}".format(self.Storage.ESURL, self.Collection, self.ObjId)
		async with self.Storage.session().request(method="POST", url=url, data=json.dumps(upsertobj), headers={'Content-Type': 'application/json'}) as resp:
			assert resp.status == 200 or resp.status == 201, "Unexpected response code: {}".format(resp.status)
			resp_json = await resp.json()
			assert resp_json["result"] == "updated" or resp_json[
				"result"] == "created", "Creating/updating was unsuccessful"

		return self.ObjId


def serialize(v):
	if isinstance(v, datetime.datetime):
		return v.timestamp()
	else:
		return v
